"""
Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict

__all__ = ["DenseNet", "densenet121", "densenet161", "densenet169", "densenet201"]

class Transition(nn.Sequential):
    """
    减少通道数, 特征图尺寸减半
    Densenet Transition Layer:
        1 × 1 conv
        2 × 2 average pool, stride 2
    """
    def __init__(self, num_input_features, num_output_features):
        super(Transition, self).__init__()
        self.norm = nn.BatchNorm2d(num_input_features)
        self.relu = nn.ReLU(inplace=True)
        self.conv = nn.Conv2d(num_input_features, num_output_features, kernel_size=1, stride=1, bias=False)
        self.pool = nn.AvgPool2d(kernel_size=2, stride=2)

class DenseLayer(nn.Module):
    def __init__(self, num_input_features, growth_rate, bn_size, drop_rate):
        super(DenseLayer, self).__init__()
        self.norm1 = nn.BatchNorm2d(num_input_features)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv2d(num_input_features, bn_size * growth_rate, kernel_size=1, stride=1, bias=False)

        self.norm2 = nn.BatchNorm2d(bn_size * growth_rate)
        self.relu2 = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(bn_size * growth_rate, growth_rate, kernel_size=3, stride=1, padding=1, bias=False)

        self.drop_rate = float(drop_rate)

    def bn_function(self, inputs):
        concated_features = torch.cat(inputs, 1)
        bottleneck_output = self.conv1(self.relu1(self.norm1(concated_features)))
        return bottleneck_output

    def forward(self, input):
        # 如果输入是张量，转化为列表，用于存储前一层的输出特征
        prev_features = [input] if torch.is_tensor(input) else input
        bottleneck_output = self.bn_function(prev_features)

        new_features = self.conv2(self.relu2(self.norm2(bottleneck_output)))
        if self.drop_rate > 0:
            new_features = F.dropout(new_features, p=self.drop_rate, training=self.training)
        return new_features

class DenseBlock(nn.Module):
    def __init__(self, num_layers, num_input_features, bn_size, growth_rate, drop_rate):
        super(DenseBlock, self).__init__()
        for i in range(num_layers):
            layer = DenseLayer(
                num_input_features + i * growth_rate,
                growth_rate=growth_rate,
                bn_size=bn_size,
                drop_rate=drop_rate,
                )
            self.add_module("denselayer%d" % (i + 1), layer)

    def forward(self, init_features):
        features = [init_features]
        for name, layer in self.named_children():
            # name为层的名称，如"denselayer1"，layer为对应的DenseLayer实例
            new_features = layer(features)
            features.append(new_features)
        return torch.cat(features, 1)

class DenseNet(nn.Module):
    def __init__(self, growth_rate=32, num_init_features=64, block_config=None, num_classes = 1000,
                 bn_size=4, drop_rate=0.):

        super(DenseNet, self).__init__()

        # First convolution
        self.features = nn.Sequential(
            OrderedDict(
                [
                    ("conv0", nn.Conv2d(3, num_init_features, kernel_size=7, stride=2, padding=3, bias=False)),
                    ("norm0", nn.BatchNorm2d(num_init_features)),
                    ("relu0", nn.ReLU(inplace=True)),
                    ("pool0", nn.MaxPool2d(kernel_size=3, stride=2, padding=1)),
                ]
            )
        )

        # Each denseblock
        num_features = num_init_features
        for i, num_layers in enumerate(block_config):
            block = DenseBlock(
                num_layers=num_layers,
                num_input_features=num_features,
                bn_size=bn_size,
                growth_rate=growth_rate,
                drop_rate=drop_rate,
            )
            self.features.add_module("denseblock%d" % (i + 1), block)
            num_features = num_features + num_layers * growth_rate
            if i != len(block_config) - 1:   # 层与层之间添加过渡层
                trans = Transition(num_input_features=num_features, num_output_features=num_features // 2)
                self.features.add_module("transition%d" % (i + 1), trans)
                num_features = num_features // 2

        # Final batch norm
        self.features.add_module("norm5", nn.BatchNorm2d(num_features))
        # Linear layer
        self.classifier = nn.Linear(num_features, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        features = self.features(x)
        out = F.relu(features, inplace=True)
        out = F.adaptive_avg_pool2d(out, (1, 1))
        out = torch.flatten(out, 1)
        out = self.classifier(out)
        return out

def densenet121(num_classes):
    """Densenet-121 model"""
    return DenseNet(32, 64, (6, 12, 24, 16),num_classes=num_classes)

def densenet161(num_classes):
    """Densenet-161 model"""
    return DenseNet(48, 96, (6, 12, 36, 24),  num_classes=num_classes)

def densenet169(num_classes):
    """Densenet-169 model"""
    return DenseNet(32, 64, (6, 12, 32, 32),   num_classes=num_classes)

def densenet201(num_classes):
    """Densenet-201 model"""
    return DenseNet(32, 64, (6, 12, 48, 32), num_classes=num_classes)


if __name__=="__main__":
    # from torchsummaryX import summary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = densenet121(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    # summary(net, torch.ones((1, 3, 224, 224)).to(device))
    # Params:    6,957.96K




class HexColors:
    def __init__(self):
        self.hexs = ('FF3838', 'FF9D97', 'FF701F', 'FFB21D', 'CFD231', '48F90A', '92CC17', '3DDB86', '1A9334', '00D4BB',
                     '2C99A8', '00C2FF', '344593', '6473FF', '0018EC', '8438FF', '520085', 'CB38FF', 'FF95C8', 'FF37C7')
        self.palette = [self.hex2rgb(f'#{c}') for c in self.hexs]
        self.n = len(self.palette)

    def __call__(self, i, bgr=False):
        c = self.palette[int(i) % self.n]
        return (c[2], c[1], c[0]) if bgr else c

    def get_hex_color(self, i):
        return self.hexs[int(i) % self.n]

    @staticmethod
    def hex2rgb(h):  # rgb order (PIL)
        return tuple(int(h[1 + i:1 + i + 2], 16) for i in (0, 2, 4))

if __name__ == "__main__":
    hex_color = '#FF0000'
    rgb_color = HexColors.hex2rgb(hex_color)
    print(f"RGB Color: {rgb_color}")
    colors = HexColors()
    index = 3
    rgb_color = colors(index)
    hex_color = colors.get_hex_color(index)
    print(f"RGB Color: {rgb_color}")
    print(f"Hex Color: {hex_color}")


"""
部分卷积 Partial convolution

Paper address: <https://arxiv.org/pdf/2303.03667.pdf>
Reference from: https://github.com/JierunChen/FasterNet/blob/master/models/fasternet.py
"""
import torch
import torch.nn as nn

class PartialConv(nn.Module):
    def __init__(self, dim, n_div, kernel_size=3, forward='split_cat'):
        """
        PartialConv 模块

        Args:
            dim (int): 输入张量的通道数。
            n_div (int): 分割通道数的分母，用于确定部分卷积的通道数。
            forward (str): 使用的前向传播方法，可选 'slicing' 或 'split_cat'。
        """
        super().__init__()
        self.dim_conv3 = dim // n_div
        self.dim_untouched = dim - self.dim_conv3
        self.partial_conv3 = nn.Conv2d(self.dim_conv3, self.dim_conv3, kernel_size, 1, 1, bias=False)

        if forward == 'slicing':
            self.forward = self.forward_slicing
        elif forward == 'split_cat':
            self.forward = self.forward_split_cat
        else:
            raise NotImplementedError

    def forward_slicing(self, x):
        # only for inference
        x = x.clone()   # !!! Keep the original input intact for the residual connection later
        x[:, :self.dim_conv3, :, :] = self.partial_conv3(x[:, :self.dim_conv3, :, :])

        return x

    def forward_split_cat(self, x):
        # for training/inference
        x1, x2 = torch.split(x, [self.dim_conv3, self.dim_untouched], dim=1)
        x1 = self.partial_conv3(x1)
        x = torch.cat((x1, x2), 1)

        return x



if __name__ == '__main__':
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    block = PartialConv(16, 4).to(device)
    input = torch.rand(1, 16, 64, 64).to(device)
    print(input.shape)
    output = block(input)
    print(output.shape)


import torch
import sys
from pyzjr.Models.backbone import *

def get_clsnetwork(name, num_classes, weights=''):
    if name == 'lenet':
        net = LeNet(num_classes=10)
    elif name == 'alexnet':
        net = AlexNet(num_classes=num_classes)
    elif name == 'vgg16':
        net = vgg16_bn(num_classes=num_classes)
    elif name == 'vgg19':
        net = vgg19_bn(num_classes=num_classes)
    elif name == 'googlenet':
        net = GoogLeNet(num_classes=num_classes)
    elif name == 'resnet18':
        net = resnet18(num_classes=num_classes)
    elif name == 'resnet34':
        net = resnet34(num_classes=num_classes)
    elif name == 'resnet50':
        net = resnet50(num_classes=num_classes)
    elif name == 'resnet101':
        net = resnet101(num_classes=num_classes)
    elif name == 'resnet152':
        net = resnet152(num_classes=num_classes)
    elif name == 'conv2former_n':
        net = Conv2Former_n(num_classes=num_classes)
    elif name == 'conv2former_t':
        net = Conv2Former_t(num_classes=num_classes)
    elif name == 'conv2former_s':
        net = Conv2Former_s(num_classes=num_classes)
    elif name == 'conv2former_l':
        net = Conv2Former_l(num_classes=num_classes)
    elif name == 'conv2former_b':
        net = Conv2Former_b(num_classes=num_classes)
    elif name == 'se_resnet18':
        net = se_resnet18(num_classes=num_classes)
    elif name == 'se_resnet34':
        net = se_resnet34(num_classes=num_classes)
    elif name == 'se_resnet50':
        net = se_resnet50(num_classes=num_classes)
    elif name == 'se_resnet101':
        net = se_resnet101(num_classes=num_classes)
    elif name == 'se_resnet152':
        net = se_resnet152(num_classes=num_classes)
    else:
        print('the network name you have entered is not supported yet')
        sys.exit()
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    net.to(device)

    if weights != '':
        print('Load weights {}.'.format(weights))
        model_dict = net.state_dict()
        pretrained_dict = torch.load(weights, map_location=device)

        matched_keys = set(model_dict.keys()) & set(pretrained_dict.keys())
        mismatched_keys = set(model_dict.keys()) - set(pretrained_dict.keys())

        for key in matched_keys:
            if model_dict[key].shape == pretrained_dict[key].shape:
                model_dict[key] = pretrained_dict[key]
            else:
                mismatched_keys.add(key)

        if mismatched_keys:
            print("The following keys have mismatched shapes or are not present in the model:")
            for key in mismatched_keys:
                print(f"- {key}")
        else:
            print("All weights were successfully loaded.")
            net.load_state_dict(model_dict)
    else:
        print("\033[31mNo weights specified.Training from scratch.But it does not affect the progress of infering")

    return net


"""
论文原址： <https://arxiv.org/pdf/1409.4842.pdf>
Going deeper with convolutions
Obtained from torchvision.models
"""
import torch
import torch.nn as nn

__all__ = ["GoogLeNet"]

class Inception(nn.Module):
    def __init__(self, in_channels, ch1x1, ch3x3red, ch3x3, ch5x5red, ch5x5, pool_proj):
        super(Inception, self).__init__()
        self.branch1 = self.BasicConv(in_channels, ch1x1, kernel_size=1)
        self.branch2 = nn.Sequential(
            self.BasicConv(in_channels, ch3x3red, kernel_size=1),
            self.BasicConv(ch3x3red, ch3x3, kernel_size=3, padding=1)
        )
        self.branch3 = nn.Sequential(
            self.BasicConv(in_channels, ch5x5red, kernel_size=1),
            self.BasicConv(ch5x5red, ch5x5, kernel_size=3, padding=1)
        )
        self.branch4 = nn.Sequential(
            nn.MaxPool2d(kernel_size=3, stride=1, padding=1, ceil_mode=True),
            self.BasicConv(in_channels, pool_proj, kernel_size=1)
        )

    def BasicConv(self, in_channels, out_channels, **kwargs):
        return nn.Sequential(nn.Conv2d(in_channels, out_channels, bias=False, **kwargs),
                             nn.BatchNorm2d(out_channels, eps=0.001),
                             nn.ReLU(inplace=True),)

    def forward(self, x):
        branch1 = self.branch1(x)
        branch2 = self.branch2(x)
        branch3 = self.branch3(x)
        branch4 = self.branch4(x)

        outputs = [branch1, branch2, branch3, branch4]
        return torch.cat(outputs, dim=1)

class GoogLeNet(nn.Module):
    def __init__(self, num_classes=1000, init_weights=False):
        super(GoogLeNet, self).__init__()
        # 可参考 https://arxiv.org/pdf/1409.4842.pdf 第七页结构图和第六页Table 1
        self.conv1 = self.BasicConv(3, 64, kernel_size=7, stride=2, padding=3)
        self.maxpool1 = nn.MaxPool2d(3, stride=2, ceil_mode=True)
        self.conv2 = self.BasicConv(64, 64, kernel_size=1, stride=1)
        self.conv3 = self.BasicConv(64, 192, kernel_size=3, padding=1)
        self.maxpool2 = nn.MaxPool2d(3, stride=2, ceil_mode=True)

        self.inception3a = Inception(192, 64, 96, 128, 16, 32, 32)
        self.inception3b = Inception(256, 128, 128, 192, 32, 96, 64)
        self.maxpool3 = nn.MaxPool2d(3, stride=2, ceil_mode=True)

        self.inception4a = Inception(480, 192, 96, 208, 16, 48, 64)
        self.inception4b = Inception(512, 160, 112, 224, 24, 64, 64)
        self.inception4c = Inception(512, 128, 128, 256, 24, 64, 64)
        self.inception4d = Inception(512, 112, 144, 288, 32, 64, 64)
        self.inception4e = Inception(528, 256, 160, 320, 32, 128, 128)
        self.maxpool4 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.inception5a = Inception(832, 256, 160, 320, 32, 128, 128)
        self.inception5b = Inception(832, 384, 192, 384, 48, 128, 128)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(0.2)
        self.fc = nn.Linear(1024, num_classes)

        if init_weights:
            self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
                import scipy.stats as stats
                X = stats.truncnorm(-2, 2, scale=0.01)
                values = torch.as_tensor(X.rvs(m.weight.numel()), dtype=m.weight.dtype)
                values = values.view(m.weight.size())
                with torch.no_grad():
                    m.weight.copy_(values)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def BasicConv(self, in_channels, out_channels, **kwargs):
        return nn.Sequential(nn.Conv2d(in_channels, out_channels, bias=False, **kwargs),
                             nn.BatchNorm2d(out_channels, eps=0.001),
                             nn.ReLU(inplace=True),)

    def forward(self, x):
        # N x 3 x 224 x 224
        x = self.conv1(x)
        # N x 64 x 112 x 112
        x = self.maxpool1(x)
        # N x 64 x 56 x 56
        x = self.conv2(x)
        # N x 64 x 56 x 56
        x = self.conv3(x)
        # N x 192 x 56 x 56
        x = self.maxpool2(x)
        # N x 192 x 28 x 28
        x = self.inception3a(x)
        # N x 256 x 28 x 28
        x = self.inception3b(x)
        # N x 480 x 28 x 28
        x = self.maxpool3(x)
        # N x 480 x 14 x 14
        x = self.inception4a(x)
        # N x 512 x 14 x 14
        x = self.inception4b(x)
        # N x 512 x 14 x 14
        x = self.inception4c(x)
        # N x 512 x 14 x 14
        x = self.inception4d(x)
        # N x 528 x 14 x 14
        x = self.inception4e(x)
        # N x 832 x 14 x 14
        x = self.maxpool4(x)
        # N x 832 x 7 x 7
        x = self.inception5a(x)
        # N x 832 x 7 x 7
        x = self.inception5b(x)
        # N x 1024 x 7 x 7
        x = self.avgpool(x)
        # N x 1024 x 1 x 1
        x = torch.flatten(x, 1)
        # N x 1024
        x = self.dropout(x)
        x = self.fc(x)
        # N x 1000 (num_classes)
        return x


if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = GoogLeNet(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # Total params: 5,604,004


"""
Copyright (c) 2024, Auorui.
All rights reserved.
http://vision.stanford.edu/cs598_spring07/papers/Lecun98.pdf
"""
import torch
import torch.nn as nn

class LeNet(nn.Module):
    def __init__(self, num_classes=10):
        super(LeNet, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=1, out_channels=6, kernel_size=5)
        self.max_pool1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv2 = nn.Conv2d(in_channels=6, out_channels=16, kernel_size=5)
        self.max_pool2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv3 = nn.Conv2d(in_channels=16, out_channels=120, kernel_size=4)
        self.fc1 = nn.Linear(120, 64)
        self.fc2 = nn.Linear(64, num_classes)

    def forward(self, x):
        x = self.conv1(x)
        x = torch.sigmoid(x)
        x = self.max_pool1(x)
        x = torch.sigmoid(x)
        x = self.conv2(x)
        x = self.max_pool2(x)
        x = self.conv3(x)
        # 输入层将数据拉平[B,C,H,W] -> [B,C*H*W]
        x = torch.reshape(x, [x.shape[0], -1])
        x = self.fc1(x)
        x = torch.sigmoid(x)
        x = self.fc2(x)
        return x


if __name__=='__main__':
    import torchsummary
    model = LeNet()
    print(model)
    torchsummary.summary(model, (1, 28, 28))
    # Total params: 41,806


"""LeNet
--------------------------------------------
Layer (type)               Output Shape
============================================
Conv2d-1                  [-1, 6, 24, 24]    
MaxPool2d-2               [-1, 6, 12, 12]          
Conv2d-3                  [-1, 16, 8, 8]         
MaxPool2d-4               [-1, 16, 4, 4]              
Conv2d-5                  [-1, 120, 1, 1]       
Linear-6                  [-1, 64]        
Linear-7                  [-1, 10]             
============================================
"""


"""
Copyright (c) 2023, Auorui.
All rights reserved.

reference <https://arxiv.org/pdf/2211.11943.pdf> (Conv2Former: A Simple Transformer-Style ConvNet for Visual Recognition)
Time:2023.12.31, Complete before the end of 2023.
"""
import torch
import torch.nn as nn

from pyzjr.Models.bricks import DropPath

__all__=["Conv2Former_n", "Conv2Former_t", "Conv2Former_s", "Conv2Former_b", "Conv2Former_l"]


class MLP(nn.Module):
    def __init__(self, dim, mlp_ratio=4.):
        super().__init__()

        self.norm = nn.LayerNorm(dim, eps=1e-6)
        self.fc1 = nn.Conv2d(dim, dim * mlp_ratio, 1)
        self.pos = nn.Conv2d(dim * mlp_ratio, dim * mlp_ratio, 3, padding=1, groups=dim * mlp_ratio)
        self.fc2 = nn.Conv2d(dim * mlp_ratio, dim, 1)
        self.act = nn.GELU()

    def forward(self, x):
        x = self.norm(x.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        x = self.fc1(x)
        x = self.act(x)
        x = x + self.act(self.pos(x))
        x = self.fc2(x)
        return x


class ConvMod(nn.Module):
    def __init__(self, dim):
        super().__init__()

        self.norm = nn.LayerNorm(dim, eps=1e-6)
        self.a = nn.Sequential(
            nn.Conv2d(dim, dim, 1),
            nn.GELU(),
            nn.Conv2d(dim, dim, 11, padding=5, groups=dim)
        )
        self.v = nn.Conv2d(dim, dim, 1)
        self.proj = nn.Conv2d(dim, dim, 1)

    def forward(self, x):
        x = self.norm(x.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        a = self.a(x)
        x = a * self.v(x)
        x = self.proj(x)
        return x



class Block(nn.Module):
    def __init__(self, dim, mlp_ratio=4.0, drop_path=0.):
        super().__init__()

        self.attn = ConvMod(dim)
        self.mlp = MLP(dim, mlp_ratio)
        layer_scale_init_value = 1e-6
        self.layer_scale_1 = nn.Parameter(layer_scale_init_value * torch.ones((dim)), requires_grad=True)
        self.layer_scale_2 = nn.Parameter(layer_scale_init_value * torch.ones((dim)), requires_grad=True)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(self, x):
        x = x + self.drop_path(self.layer_scale_1.unsqueeze(-1).unsqueeze(-1) * self.attn(x))
        x = x + self.drop_path(self.layer_scale_2.unsqueeze(-1).unsqueeze(-1) * self.mlp(x))
        return x

class BaseLayer(nn.Module):
    def __init__(self, dim, depth, mlp_ratio=4., drop_path=None, downsample=True):
        super().__init__()
        self.dim = dim
        self.drop_path = drop_path

        self.blocks = nn.ModuleList([
            Block(dim=self.dim,mlp_ratio=mlp_ratio,drop_path=drop_path[i],)
            for i in range(depth)
        ])

        # patch merging layer
        if downsample:
            self.downsample = nn.Sequential(
                nn.GroupNorm(num_groups=1, num_channels=dim),
                nn.Conv2d(dim, dim * 2, kernel_size=2, stride=2,bias=False)
            )
        else:
            self.downsample = None

    def forward(self, x):
        for blk in self.blocks:
            x = blk(x)
        if self.downsample is not None:
            x = self.downsample(x)
        return x

class Conv2Former(nn.Module):
    def __init__(self, num_classes=10, depths=(2,2,8,2), dim=(64,128,256,512), mlp_ratio=2.,drop_rate=0.,
                 drop_path_rate=0.15, **kwargs):
        super().__init__()

        norm_layer = nn.LayerNorm
        self.num_classes = num_classes
        self.num_layers = len(depths)
        self.dim = dim
        self.mlp_ratio = mlp_ratio
        self.pos_drop = nn.Dropout(p=drop_rate)

        # stochastic depth decay rule
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]

        # build layers
        self.layers = nn.ModuleList()
        for i_layer in range(self.num_layers):
            layer = BaseLayer(dim[i_layer],
                              depth=depths[i_layer],
                              mlp_ratio=self.mlp_ratio,
                              drop_path=dpr[sum(depths[:i_layer]):sum(depths[:i_layer + 1])],
                              downsample=(i_layer < self.num_layers - 1),
                              )
            self.layers.append(layer)
        self.fc1 = nn.Conv2d(3, dim[0], 1)
        self.norm = norm_layer(dim[-1], eps=1e-6,)
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.head = nn.Linear(dim[-1], num_classes) \
            if num_classes > 0 else nn.Identity()

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            if m.bias is not None:
                nn.init.constant_(m.bias, 0.)
        elif isinstance(m, (nn.Conv1d, nn.Conv2d)):
            nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            if m.bias is not None:
                nn.init.constant_(m.bias, 0.)
        elif isinstance(m, (nn.LayerNorm, nn.GroupNorm)):
            nn.init.constant_(m.bias, 0.)
            nn.init.constant_(m.weight, 1.)

    def forward_features(self, x):
        x = self.fc1(x)
        x = self.pos_drop(x)

        for layer in self.layers:
            x = layer(x)

        x = self.norm(x.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        return x

    def forward(self, x):
        x = self.forward_features(x)
        x = self.head(x)
        return x


#  reference <https://arxiv.org/pdf/2211.11943.pdf> Table 1
C = {'n': [64, 128, 256, 512],
     't': [72, 144, 288, 576],
     's': [72, 144, 288, 576],
     'b': [96, 192, 384, 768],
     'l': [128, 256, 512, 1024],
     }
L = {'n': [2, 2, 8, 2],
     't': [3, 3, 12, 3],
     's': [4, 4, 32, 4],
     'b': [4, 4, 34, 4],
     'l': [4, 4, 48, 4],
     }

def Conv2Former_n(num_classes, mlp_ratio=2, drop_path_rate=0.1):
    model = Conv2Former(num_classes=num_classes, depths=L["n"], dim=C["n"], mlp_ratio=mlp_ratio, drop_path_rate=drop_path_rate)
    return model

def Conv2Former_t(num_classes, mlp_ratio=2, drop_path_rate=0.1):
    model = Conv2Former(num_classes=num_classes, depths=L["t"], dim=C["t"], mlp_ratio=mlp_ratio, drop_path_rate=drop_path_rate)
    return model

def Conv2Former_s(num_classes, mlp_ratio=2, drop_path_rate=0.1):
    model = Conv2Former(num_classes=num_classes, depths=L["s"], dim=C["s"], mlp_ratio=mlp_ratio, drop_path_rate=drop_path_rate)
    return model

def Conv2Former_b(num_classes, mlp_ratio=2, drop_path_rate=0.1):
    model = Conv2Former(num_classes=num_classes, depths=L["b"], dim=C["b"], mlp_ratio=mlp_ratio, drop_path_rate=drop_path_rate)
    return model

def Conv2Former_l(num_classes, mlp_ratio=2, drop_path_rate=0.1):
    model = Conv2Former(num_classes=num_classes, depths=L["l"], dim=C["l"], mlp_ratio=mlp_ratio, drop_path_rate=drop_path_rate)
    return model

if __name__ == '__main__':
    import torchsummary
    model = Conv2Former_n(num_classes=4)
    input_tensor = torch.ones(2, 3, 224, 224).cpu()
    model = model.cpu()
    output = model(input_tensor)
    print("Output shape:", output.shape)
    torchsummary.summary(model, input_size=(3, 224, 224))
    # n: Total params: 8,844,420
    # t: Total params: 16,241,044
    # 其他的结构容易导致CUDA error: out of memory,建议使用前面两个类型，即是 n 和 t


"""
mediapipe
- author: Auorui (夏天是冰红茶)
- creation time:2022.5
- pyzjr.mediapipe is a part of the model strategy used for deep learning
- WeChat: z15583909992
- Email: zjricetea@gmail.com
- Note: Currently still being updated, please refer to the latest version for any changes that may occur
"""


from .HandTrack import Hands


__all__ = ["Hands",]


from .general import is_pil, is_rgb_image, is_numpy, is_tensor, is_gray_image
import numpy as np

__all__=["not_rgb_warning","Error_flags","Augment_Error_Message", "_check_img_is_plt", "_check_parameter_is_tuple_2", "_check_parameter_is_tuple_and_list_2", \
         "_check_parameter_is_tuple_and_list_or_single_2", "check_dtype", "_check_img_is_ndarray", "_check_input_is_tensor" ,
         "_get_image_num_channels"]
Error_flags = "[pyzjr error]:"

Augment_Error_Message = "Image should be PIL image. Got {}. Use Decode() for encoded data or ToPIL() for decoded data."
Ndarray_Error_Message = "The input image should be of type numpy.ndarray or PIL.Image.Image. Got {}."

def not_rgb_warning(image):
    if not is_rgb_image(image):
        message = "This transformation expects 3-channel images"
        if is_gray_image(image):
            message += "\nYou can convert your gray image to RGB using cv2.cvtColor(image, cv2.COLOR_GRAY2RGB))"
        raise ValueError(message)

def _check_img_is_plt(img):
    if not is_pil(img):
        raise TypeError(Augment_Error_Message.format(type(img)))

def _check_img_is_ndarray(img):
    if not is_numpy(img):
        raise TypeError(Ndarray_Error_Message.format(type(img)))

def _check_img_is_opencv(opencv_img):
    if isinstance(opencv_img, np.ndarray):
        return True
    else:
        raise ValueError("The loaded image should conform to the format read by OpenCV!")

def _check_input_is_tensor(img):
    if not is_tensor(img):
        raise TypeError('Input tensor should be a torch tensor. Got {}.'.format(type(img)))

def _check_parameter_is_tuple_2(a):
    if not isinstance(a, tuple) or len(a) != 2:
        raise ValueError(f"{a} should be a tuple (x_para,y_para).")

def _check_parameter_is_tuple_and_list_2(a):
    if not isinstance(a, (list, tuple)) or len(a) != 2:
        raise ValueError(f"{a} should be a list or tuple of two values.")

def _check_parameter_is_tuple_and_list_or_single_2(a):
    if not (isinstance(a, int) or (isinstance(a, (list, tuple)) and len(a) == 2)):
        raise TypeError('Should be a single number or a list/tuple (h, w) of length 2.'
                        'Got {}.'.format(a))

def convert_to_tuple(value):
    if isinstance(value, int):
        return (value, value)
    elif isinstance(value, (list, tuple)):
        if len(value) == 2:
            return value
        else:
            raise ValueError("The input object is of type list or tuple, but the length is not two!")

def _get_image_num_channels(img):
    if is_tensor(img):
        if img.ndim == 2:
            return 1
        elif img.ndim > 2:
            return img.shape[-3]
    if is_pil(img):
        return 1 if img.mode == 'L' else 3
    if is_numpy(img):
        return 1 if is_gray_image else 3


class check_dtype():
    def __init__(self, para):
        self.para = para

    def is_int(self):
        return isinstance(self.para, int)

    def is_float(self):
        return isinstance(self.para, float)

    def is_list(self):
        return isinstance(self.para, list)

    def is_tuple(self):
        return isinstance(self.para, tuple)

    def is_list_or_tuple(self):
        return isinstance(self.para, (list, tuple))

    def is_None(self):
        return self.para is None

    def is_positive_int(self):
        return self.is_int() and self.para > 0

    def is_nonnegative_int(self):
        return self.is_int() and self.para >= 0



"""
Copyright (c) 2024, Auorui.
All rights reserved.

Weighted average, reference from <https://pytorch.org/docs/stable/optim.html#putting-it-all-together-ema>
指数移动平均（EMA）   常见
随机加权平均（SWA）   https://arxiv.org/abs/1803.05407
Tanh自适应指数移动EMA算法（T_ADEMA）    Originating from https://kns.cnki.net/kcms2/article/abstract?v=vdPasdvfHvuuN-gB4G4neXcKqbiX3vnPwH2QfVTNb4OQCyQJO2HTgHpa6C3EDWqlnkrcNyjnTtxYjmozsntFZzru-e7vk_X4Fq9NCmtavFoJFkztVbWX1vr5qj9w2djSGdHx0-RWLb0=&uniplatform=NZKPT&flag=copy
https://blog.csdn.net/m0_62919535/article/details/135482009?spm=1001.2014.3001.5501
time 2024-01-09
"""
import torch
from copy import deepcopy
from pyzjr.core.general import is_parallel
import itertools
from torch.nn import Module

__all__ = ["AveragingBaseModel", "EMAModel", "SWAModel", "T_ADEMAModel"]

def de_parallel(model):
    """
    将并行模型（DataParallel 或 DistributedDataParallel）转换为单 GPU 模型。
    """
    return model.module if is_parallel(model) else model

def get_ema_avg_fn(decay=0.999):
    @torch.no_grad()
    def ema_update(ema_param, current_param, num_averaged):
        return decay * ema_param + (1 - decay) * current_param
    return ema_update

def get_swa_avg_fn():
    @torch.no_grad()
    def swa_update(averaged_param, current_param, num_averaged):
        return averaged_param + (current_param - averaged_param) / (num_averaged + 1)
    return swa_update

def get_t_adema_fn(alpha=0.9):
    num_averaged = [0]  # 使用列表包装可变对象，以在闭包中引用
    @torch.no_grad()
    def t_adema_update(averaged_param, current_param, num_averageds):
        num_averaged[0] += 1
        decay = alpha * torch.tanh(torch.tensor(num_averaged[0], dtype=torch.float32))
        tadea_update = decay * averaged_param + (1 - decay) * current_param
        return tadea_update
    return t_adema_update


class AveragingBaseModel(Module):
    def __init__(self, model, cuda=False, avg_fn=None, use_buffers=False):
        super(AveragingBaseModel, self).__init__()
        device = 'cuda' if cuda and torch.cuda.is_available() else 'cpu'
        self.module = deepcopy(de_parallel(model))
        self.module = self.module.to(device)
        self.register_buffer('n_averaged',
                             torch.tensor(0, dtype=torch.long, device=device))
        self.avg_fn = avg_fn
        self.use_buffers = use_buffers

    def forward(self, *args, **kwargs):
        return self.module(*args, **kwargs)

    def update(self, model):
        self_param = itertools.chain(self.module.parameters(), self.module.buffers() if self.use_buffers else [])
        model_param = itertools.chain(model.parameters(), model.buffers() if self.use_buffers else [])

        self_param_detached = [p.detach() for p in self_param]
        model_param_detached = [p.detach().to(p_averaged.device) for p, p_averaged in zip(model_param, self_param_detached)]

        if self.n_averaged == 0:
            for p_averaged, p_model in zip(self_param_detached, model_param_detached):
                p_averaged.copy_(p_model)

        if self.n_averaged > 0:
            for p_averaged, p_model in zip(self_param_detached, model_param_detached):
                n_averaged = self.n_averaged.to(p_averaged.device)
                p_averaged.copy_(self.avg_fn(p_averaged, p_model, n_averaged))

        if not self.use_buffers:
            for b_swa, b_model in zip(self.module.buffers(), model.buffers()):
                b_swa.copy_(b_model.to(b_swa.device).detach())

        self.n_averaged += 1

class EMAModel(AveragingBaseModel):
    def __init__(self, model, cuda = False, decay=0.99, use_buffers=False):
        super().__init__(model=model, cuda=cuda, avg_fn=get_ema_avg_fn(decay), use_buffers=use_buffers)

class SWAModel(AveragingBaseModel):
    def __init__(self, model, cuda = False,use_buffers=False):
        super().__init__(model=model, cuda=cuda, avg_fn=get_swa_avg_fn(), use_buffers=use_buffers)

class T_ADEMAModel(AveragingBaseModel):
    def __init__(self, model, cuda=False, alpha=0.99, use_buffers=False):
        super().__init__(model=model, cuda=cuda, avg_fn=get_t_adema_fn(alpha), use_buffers=use_buffers)


if __name__=="__main__":
    # 创建 ResNet18 模型
    import torchvision.models as models
    from torch.utils.data import DataLoader
    from tqdm import tqdm

    class RandomDataset(torch.utils.data.Dataset):
        def __init__(self, size=224):
            self.data = torch.randn(size, 3, 224, 224)
            self.labels = torch.randint(0, 2, (size,))

        def __getitem__(self, index):
            return self.data[index], self.labels[index]

        def __len__(self):
            return len(self.data)


    model = models.resnet18(pretrained=False)
    # model = model.to('cuda')
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
    criterion = torch.nn.CrossEntropyLoss()

    # 创建数据加载器
    train_dataset = RandomDataset()
    train_loader = DataLoader(train_dataset, batch_size=32, shuffle=True)

    # 定义权重平均模型
    swa_model = SWAModel(model, cuda=True)
    ema_model = EMAModel(model, cuda=True)
    t_adema_model = T_ADEMAModel(model, cuda=True)

    for epoch in range(5):
        for inputs, labels in tqdm(train_loader, desc=f"Epoch {epoch + 1}/{5}"):
            optimizer.zero_grad()
            outputs = model(inputs)
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()

            # 更新权重平均模型
            ema_model.update(model)
            swa_model.update(model)
            t_adema_model.update(model)

    # 测试模型
    test_dataset = RandomDataset(size=100)
    test_loader = DataLoader(test_dataset, batch_size=32, shuffle=False)


    def evaluate(model):
        model.eval()  # 切换到评估模式
        correct = 0
        total = 0

        with torch.no_grad():
            for inputs, labels in test_loader:
                inputs, labels = inputs.to('cuda'), labels.to('cuda')
                outputs = model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        accuracy = correct / total
        print(f"模型准确率：{accuracy * 100:.2f}%")

    # 原模型测试
    print("Model Evaluation:")
    evaluate(model.to('cuda'))   #
    # 测试权重平均模型
    print("SWAModel Evaluation:")
    evaluate(swa_model.to('cuda'))

    print("EMAModel Evaluation:")
    evaluate(ema_model.to('cuda'))

    print("T-ADEMAModel Evaluation:")
    evaluate(t_adema_model.to('cuda'))

    # 仅仅测试是否能够跑通，过程中也有比元模型要低的时候，而且权值平均主要是用与训练中后期
    # Model Evaluation:
    # 模型准确率：46.00 %
    # SWAModel Evaluation:
    # 模型准确率：54.00% %
    # EMAModel Evaluation:
    # 模型准确率：58.00% %
    # T - ADEMAModel Evaluation:
    # 模型准确率：58.00% %



"""
Time: 2024-01-28 0:59
"""
import torch
import math
import torch.nn as nn
import torch.nn.functional as F

def conv3x3(in_channels, out_channels, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding:捕捉局部特征和空间相关性，学习更复杂的特征和抽象表示"""
    return nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)

def conv1x1(in_channels, out_channels, stride=1):
    """1x1 convolution:实现降维或升维，调整通道数和执行通道间的线性变换"""
    return nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)

class ConvBNReLU(nn.Module):
    def __init__(self, in_channels, out_channels, stride=1,  kernel_size=3, padding=1,
                 **kwargs):
        super(ConvBNReLU, self).__init__()
        self.conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels,
                              kernel_size=kernel_size, stride=stride, padding=padding,
                              **kwargs)
        self.bn = nn.BatchNorm2d(num_features=out_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x

class ConvBNActivation(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, groups=1,
                 norm_layer=None, activation_layer = None, dilation=1,):
        super(ConvBNActivation, self).__init__()
        padding = (kernel_size - 1) // 2 * dilation
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if activation_layer is None:
            activation_layer = nn.ReLU6
        self.convbnact=nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, dilation=dilation, groups=groups,
                      bias=False),
            norm_layer(out_planes),
            activation_layer(inplace=True)
        )
        self.out_channels = out_planes

    def forward(self, x):
        return self.convbnact(x)

class ConvBN(nn.Module):
    def __init__(self, in_channels, out_channels, stride=1,  kernel_size=3, padding=1,
                 **kwargs):
        super(ConvBN, self).__init__()
        self.conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels,
                              kernel_size=kernel_size, stride=stride, padding=padding,
                              **kwargs)
        self.bn = nn.BatchNorm2d(num_features=out_channels)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        return x

class BasicBlock(nn.Module):
    """ResNet的基础块,适用于较浅的网络或较小的数据集"""
    expansion = 1
    def __init__(self, in_channels, out_channels, stride=1, downsample=False):
        super(BasicBlock, self).__init__()
        self.convbnrelu = ConvBNReLU(in_channels, out_channels, kernel_size=3, stride=stride)
        self.convbn = ConvBN(out_channels, out_channels, kernel_size=3)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.conv_down = nn.Sequential(
            conv1x1(in_channels, out_channels * self.expansion, self.stride),
            nn.BatchNorm2d(out_channels * self.expansion),
        )

    def forward(self, x):
        residual = x
        out = self.convbnrelu(x)
        out = self.convbn(out)

        if self.downsample:
            residual = self.conv_down(x)

        out += residual
        out = self.relu(out)

        return out

class Bottleneck(nn.Module):
    """ResNet的瓶颈结构,适用于深层网络，通过较少的参数来构建更深的网络，提高性能"""
    expansion = 4
    def __init__(self, in_channels, out_channels, stride=1, downsample=False):
        super(Bottleneck, self).__init__()
        groups = 1
        base_width = 64
        dilation = 1

        width = int(out_channels * (base_width / 64.)) * groups   # wide = out_channels
        # self.convbnrelu1 = ConvBNReLU(in_channels, width, kernel_size=1, padding=0)  # 降维通道数
        # self.convbnrelu2 = ConvBNReLU(width, width, kernel_size=3, stride=stride, dilation=dilation, groups=groups)
        # self.convbn3 = ConvBN(width, out_channels * self.expansion, kernel_size=1, padding=0)   # 升维通道数
        self.conv1 = conv1x1(in_channels, width)       # 降维通道数
        self.bn1 = nn.BatchNorm2d(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = nn.BatchNorm2d(width)
        self.conv3 = conv1x1(width, out_channels * self.expansion)   # 升维通道数
        self.bn3 = nn.BatchNorm2d(out_channels * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.conv_down = nn.Sequential(
            conv1x1(in_channels, out_channels * self.expansion, self.stride),
            nn.BatchNorm2d(out_channels * self.expansion),
        )

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        # out = self.convbnrelu1(x)
        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)
        # out = self.convbnrelu2(out)
        out = self.conv3(out)
        out = self.bn3(out)
        # out = self.convbn3(out)
        if self.downsample:
            residual = self.conv_down(x)

        out += residual
        out = self.relu(out)

        return out

class FireModule(nn.Module):
    """对通道的压缩和扩展(SqueezeNet)"""
    def __init__(self, inplanes, squeeze_planes, expand1x1_planes, expand3x3_planes):
        super(FireModule, self).__init__()
        self.inplanes = inplanes
        self.squeeze = nn.Conv2d(inplanes, squeeze_planes, kernel_size=1)
        self.expand1x1 = nn.Conv2d(squeeze_planes, expand1x1_planes,
                                   kernel_size=1)
        self.expand3x3 = nn.Conv2d(squeeze_planes, expand3x3_planes,
                                   kernel_size=3, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.relu(self.squeeze(x))
        return torch.cat([
            self.relu(self.expand1x1(x)),
            self.relu(self.expand3x3(x))
        ], dim=1)

class _DenseLayer(nn.Sequential):
    def __init__(self, num_input_features, growth_rate, bn_size, drop_rate):
        super(_DenseLayer, self).__init__()
        self.add_module('norm1', nn.BatchNorm2d(num_input_features)),
        self.add_module('relu1', nn.ReLU(inplace=True)),
        self.add_module('conv1', nn.Conv2d(num_input_features, bn_size *
                                           growth_rate, kernel_size=(1,1), stride=(1,1), bias=False)),
        self.add_module('norm2', nn.BatchNorm2d(bn_size * growth_rate)),
        self.add_module('relu2', nn.ReLU(inplace=True)),
        self.add_module('conv2', nn.Conv2d(bn_size * growth_rate, growth_rate,
                                           kernel_size=3, stride=(1,1), padding=1, bias=False)),
        self.drop_rate = drop_rate

    def forward(self, x):
        new_features = super(_DenseLayer, self).forward(x)
        if self.drop_rate > 0:
            new_features = F.dropout(new_features, p=self.drop_rate, training=self.training)
        return torch.cat([x, new_features], 1)


class _DenseBlock(nn.Sequential):
    def __init__(self, num_layers, num_input_features, bn_size, growth_rate, drop_rate):
        super(_DenseBlock, self).__init__()
        for i in range(num_layers):
            layer = _DenseLayer(num_input_features + i * growth_rate, growth_rate, bn_size, drop_rate)
            self.add_module('denselayer%d' % (i + 1), layer)

class Transition(nn.Sequential):
    """
    减少通道数, 特征图尺寸减半, Densenet过渡层
    Densenet Transition Layer:
        1 × 1 conv
        2 × 2 average pool, stride 2
    """
    def __init__(self, num_input_features, num_output_features):
        super(Transition, self).__init__()
        self.norm = nn.BatchNorm2d(num_input_features)
        self.relu = nn.ReLU(inplace=True)
        self.conv = nn.Conv2d(num_input_features, num_output_features, kernel_size=1, stride=1, bias=False)
        self.pool = nn.AvgPool2d(kernel_size=2, stride=2)


class InvertedResidualv2(nn.Module):
    """MobileNetV2的倒残差结构"""
    def __init__(self, in_planes, out_planes, stride, expand_ratio):
        super(InvertedResidualv2, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        hidden_dim = int(round(in_planes * expand_ratio))
        self.use_res_connect = self.stride == 1 and in_planes == out_planes

        layers = []
        if expand_ratio != 1:
            # pw 利用1x1卷积进行通道数的上升
            layers.append(ConvBNActivation(in_planes, hidden_dim, kernel_size=1))
        layers.extend([
            # dw 进行3x3的逐层卷积，进行跨特征点的特征提取
            ConvBNActivation(hidden_dim, hidden_dim, stride=stride, groups=hidden_dim),
            # pw-linear 利用1x1卷积进行通道数的下降
            nn.Conv2d(hidden_dim, out_planes, kernel_size=1, stride=1, padding=0),
            nn.BatchNorm2d(out_planes),
        ])
        self.conv = nn.Sequential(*layers)
        self.out_channels = out_planes

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class GhostModulev1(nn.Module):
    def __init__(self, inp, oup, kernel_size=1, ratio=2, dw_size=3, stride=1, relu=True):
        super(GhostModulev1, self).__init__()
        self.oup = oup
        init_channels = math.ceil(oup / ratio)   # m = n / s
        new_channels = init_channels*(ratio-1)   # m * (s - 1) = n / s * (s - 1)

        # 利用1x1卷积对输入进来的特征图进行通道的浓缩, 实现跨通道的特征提取
        self.primary_conv = nn.Sequential(
            nn.Conv2d(inp, init_channels, kernel_size, stride, kernel_size//2, bias=False),
            nn.BatchNorm2d(init_channels),
            nn.ReLU(inplace=True) if relu else nn.Sequential(),
        )
        # 使用逐层卷积, 获得额外的特征图
        self.cheap_operation = nn.Sequential(
            nn.Conv2d(init_channels, new_channels, dw_size, stride=1, padding=dw_size//2, groups=init_channels, bias=False),
            nn.BatchNorm2d(new_channels),
            nn.ReLU(inplace=True) if relu else nn.Sequential(),
        )

    def forward(self, x):
        x1 = self.primary_conv(x)
        x2 = self.cheap_operation(x1)
        out = torch.cat([x1, x2], dim=1)
        return out[:, :self.oup, :, :]


class GhostBottleneckv1(nn.Module):
    """ Ghost bottleneck w/removed SE """
    def __init__(self, in_chs, mid_chs, out_chs, dw_size=3, stride=1):
        super(GhostBottleneckv1, self).__init__()
        self.stride = stride
        # Point-wise expansion
        self.ghost1 = GhostModulev1(in_chs, mid_chs, relu=True)
        # Depth-wise convolution
        if self.stride > 1:
            self.conv_dw = nn.Conv2d(mid_chs, mid_chs, dw_size, stride=stride,
                                     padding=(dw_size-1)//2,
                                     groups=mid_chs, bias=False)
            self.bn_dw = nn.BatchNorm2d(mid_chs)
        # Point-wise linear projection
        self.ghost2 = GhostModulev1(mid_chs, out_chs, relu=False)
        # shortcut
        if (in_chs == out_chs and self.stride == 1):
            self.shortcut = nn.Sequential()
        else:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_chs, in_chs, dw_size, stride=stride,
                          padding=(dw_size-1)//2, groups=in_chs, bias=False),
                nn.BatchNorm2d(in_chs),
                nn.Conv2d(in_chs, out_chs, 1, stride=1, padding=0, bias=False),
                nn.BatchNorm2d(out_chs),
            )

    def forward(self, x):
        residual = x
        # 1st ghost bottleneck
        x = self.ghost1(x)
        # Depth-wise convolution
        if self.stride > 1:
            x = self.conv_dw(x)
            x = self.bn_dw(x)
        # 2nd ghost bottleneck
        x = self.ghost2(x)
        x += self.shortcut(residual)
        return x


class GhostModulev2(nn.Module):
    """添加了DFC注意力机制,即short_conv部分,原文使用的是max_pool,这里使用的是avg_pool"""
    def __init__(self, inp, oup, kernel_size=1, ratio=2, dw_size=3, stride=1, relu=True, mode=None):
        super(GhostModulev2, self).__init__()
        self.mode = mode
        self.gate_fn = nn.Sigmoid()

        if self.mode in ['original']:
            self.oup = oup
            init_channels = math.ceil(oup / ratio)
            new_channels = init_channels*(ratio-1)
            self.primary_conv = nn.Sequential(
                nn.Conv2d(inp, init_channels, kernel_size, stride, kernel_size//2, bias=False),
                nn.BatchNorm2d(init_channels),
                nn.ReLU(inplace=True) if relu else nn.Sequential(),
            )
            self.cheap_operation = nn.Sequential(
                nn.Conv2d(init_channels, new_channels, dw_size, 1, dw_size//2, groups=init_channels, bias=False),
                nn.BatchNorm2d(new_channels),
                nn.ReLU(inplace=True) if relu else nn.Sequential(),
            )
        elif self.mode in ['attn']:
            self.oup = oup
            init_channels = math.ceil(oup / ratio)
            new_channels = init_channels*(ratio-1)
            self.primary_conv = nn.Sequential(
                nn.Conv2d(inp, init_channels, kernel_size, stride, kernel_size//2, bias=False),
                nn.BatchNorm2d(init_channels),
                nn.ReLU(inplace=True) if relu else nn.Sequential(),
            )
            self.cheap_operation = nn.Sequential(
                nn.Conv2d(init_channels, new_channels, dw_size, 1, dw_size//2, groups=init_channels, bias=False),
                nn.BatchNorm2d(new_channels),
                nn.ReLU(inplace=True) if relu else nn.Sequential(),
            )
            self.short_conv = nn.Sequential(
                # horizontal DFC and vertical DFC
                nn.Conv2d(inp, oup, kernel_size, stride, kernel_size//2, bias=False),
                nn.BatchNorm2d(oup),
                nn.Conv2d(oup, oup, kernel_size=(1, 5), stride=1, padding=(0,2), groups=oup,bias=False),
                nn.BatchNorm2d(oup),
                nn.Conv2d(oup, oup, kernel_size=(5, 1), stride=1, padding=(2,0), groups=oup,bias=False),
                nn.BatchNorm2d(oup),
            )

    def forward(self, x):
        if self.mode in ['original']:
            x1 = self.primary_conv(x)
            x2 = self.cheap_operation(x1)
            out = torch.cat([x1, x2], dim=1)
            return out[:, : self.oup, :, :]
        elif self.mode in ['attn']:
            res = self.short_conv(F.avg_pool2d(x, kernel_size=2, stride=2))
            x1 = self.primary_conv(x)
            x2 = self.cheap_operation(x1)
            out = torch.cat([x1, x2], dim=1)
            return out[:, :self.oup, :, :] * F.interpolate(self.gate_fn(res), size=(out.shape[-2], out.shape[-1]), mode='nearest')


class GhostBottleneckv2(nn.Module):
    """与原文有所出入,在消融实验中每层都添加了DFC,这里前两层还是用的原本v1版本"""
    def __init__(self, in_chs, mid_chs, out_chs, dw_kernel_size=3,
                 stride=1, layer_id=None):
        super(GhostBottleneckv2, self).__init__()
        self.stride = stride

        # Point-wise expansion
        if layer_id <= 1:
            self.ghost1 = GhostModulev2(in_chs, mid_chs, relu=True, mode='original')
        else:
            self.ghost1 = GhostModulev2(in_chs, mid_chs, relu=True, mode='attn')

        # Depth-wise convolution
        if self.stride > 1:
            self.conv_dw = nn.Conv2d(mid_chs, mid_chs, dw_kernel_size, stride=stride,
                                     padding=(dw_kernel_size-1)//2, groups=mid_chs, bias=False)
            self.bn_dw = nn.BatchNorm2d(mid_chs)

        self.ghost2 = GhostModulev2(mid_chs, out_chs, relu=False,mode='original')

        # shortcut
        if (in_chs == out_chs and self.stride == 1):
            self.shortcut = nn.Sequential()
        else:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_chs, in_chs, dw_kernel_size, stride=stride,
                          padding=(dw_kernel_size-1)//2, groups=in_chs, bias=False),
                nn.BatchNorm2d(in_chs),
                nn.Conv2d(in_chs, out_chs, 1, stride=1, padding=0, bias=False),
                nn.BatchNorm2d(out_chs),
            )

    def forward(self, x):
        residual = x
        x = self.ghost1(x)
        if self.stride > 1:
            x = self.conv_dw(x)
            x = self.bn_dw(x)
        x = self.ghost2(x)
        x += self.shortcut(residual)
        return x






import os
import torch

def save_models(model, epoch, save_dir, save_period = 10):
    """Save model to given path
    Args:
        model: model to be saved
        save_dir: path that the model would be saved
        epoch: the epoch the model finished training
    """
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    if epoch % save_period == 0:
        torch.save(model.state_dict(), os.path.join(save_dir, f"model_epoch_{epoch}.pth"))

_torch_save = torch.save  # copy to avoid recursion errors

def torch_save(*args, **kwargs):
    """Use dill (if exists) to serialize the lambda functions where pickle does not do this.

    Args:
        *args (tuple): Positional arguments to pass to torch.save.
        **kwargs (dict): Keyword arguments to pass to torch.save.
    """
    try:
        import dill as pickle  # noqa
    except ImportError:
        import pickle
    if 'pickle_module' not in kwargs:
        kwargs['pickle_module'] = pickle  # noqa
    return _torch_save(*args, **kwargs)



import numpy as np
import cv2
import string
from skimage.morphology import skeletonize
from skimage.filters import threshold_otsu
from skimage.color import rgb2gray
from pyzjr.core import is_gray_image

__all__=["SkeletonMap", "incircle", "outcircle"]

def SkeletonMap(target):
    """
    获取骨架图的信息
    :param target: 目标图
    :return: 骨架图与一个数组，其中每一行表示一个非零元素的索引(y,x)，包括行索引和列索引
    """
    gray = rgb2gray(target)
    thresh = threshold_otsu(gray)
    binary = gray > thresh
    skimage = skeletonize(binary)
    skepoints = np.argwhere(skimage)
    skimage = skimage.astype(np.uint8)
    return skimage, skepoints


def incircle(img, contours_arr, color=(0, 0, 255)):
    """
    轮廓最大内切圆算法,所有轮廓当中的内切圆
    :param img: 单通道图像
    :param contours_arr: ndarry的轮廓, 建议使用cv2.findContours,
                        contours_arr, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
                        pyzjr也有SearchOutline可用,但要转换成ndarry的格式
    :example:
        contour = [np.array([point], dtype=np.int32) for point in contours]
        # 平铺的方法
        flatten_contours = np.concatenate([cnt.flatten() for cnt in contours_arr])
        flatten_contours = flatten_contours.reshape(-1, 2)
    :param color: 绘制内切圆颜色
    :return: 绘制在原图的内切圆,内切圆直径,绘制出的图像与轮廓直接差距一个像素，是因为cv2.circle的半径参数必须为int类型
    """
    if is_gray_image:
        result = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    raw_dist = np.zeros(img.shape, dtype=np.float32)
    letters = list(string.ascii_uppercase)
    label = {}
    for k, contours in enumerate(contours_arr):
        for i in range(img.shape[0]):
            for j in range(img.shape[1]):
                raw_dist[i, j] = cv2.pointPolygonTest(contours, (j, i), True)
        min_val, max_val, _, max_dist_pt = cv2.minMaxLoc(raw_dist)
        label[letters[k]] = max_val * 2
        radius = int(max_val)
        cv2.circle(result, max_dist_pt, radius, color, 1, 1, 0)

    return result, label


def outcircle(img, contours_arr, color=(0,255,0)):
    """
    轮廓外切圆算法,画出所有轮廓的外切圆
    :param img: 单通道图像
    :param contours_arr: ndarry的轮廓, 建议使用cv2.findContours,
                        contours_arr, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
                        pyzjr也有SearchOutline可用,但要转换成ndarry的格式
    :example:
        contour = [np.array([point], dtype=np.int32) for point in contours]
    :param color: 绘制外切圆颜色
    :return:
    """
    radii = []
    result = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    for k, cnt in enumerate(contours_arr):
        (x, y), radius = cv2.minEnclosingCircle(cnt)
        radii.append([(x,y),radius])
        center = (int(x), int(y))
        radius = int(radius)
        cv2.circle(result, center, radius, color, 1)
        cv2.circle(result, center, 1, color, 1)

    return result, radii


class ToMask():
    """集成不同处理掩膜、标签的方式"""
    def __init__(self, mask):
        self.mask = mask

    def Binary(self, mask):
        pass

    def up_low(self, lower, upper):
        """
        :param lower: 颜色范围的下限值，作为一个包含(B, G, R)值的元组
        :param upper: 颜色范围的上限值，作为一个包含(B, G, R)值的元组
        """
        lower = np.array(lower, dtype=np.uint8)
        upper = np.array(upper, dtype=np.uint8)
        mask = cv2.inRange(self.mask, lower, upper)

        return mask

    def threshold(self, std=127.5):
        """阈值法"""
        self.mask[self.mask > std] = 255
        self.mask[self.mask < std] = 0
        mask = self.mask.astype("uint8")

        return mask

    def combine(self, operation='and', *masks):
        """
        :param operation: 组合操作，可选 'and', 'or', 'not'
        :param masks: 要组合的掩膜列表
        """
        if not masks:
            return

        combined_mask = self.mask.copy()

        for mask in masks:
            if operation == 'and':
                combined_mask = cv2.bitwise_and(combined_mask, mask)
            elif operation == 'or':
                combined_mask = cv2.bitwise_or(combined_mask, mask)
            elif operation == 'not':
                combined_mask = cv2.bitwise_not(mask)
            else:
                raise ValueError("Invalid operation. Use 'and', 'or', or 'not'.")

        return combined_mask






"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is applied to image reading and display.
"""
import os
import cv2
import imghdr
import numpy as np
from urllib import request
import torch
from matplotlib import pyplot as plt
from pyzjr.data.FM import is_url
plt.rcParams['font.sans-serif'] = ['SimHei']
plt.rcParams['axes.unicode_minus'] = False
from pathlib import Path
from pyzjr.core import is_gray_image, is_pil

__all__ = ["ImagesReader", "imwrite", "imshowplt", "StackedCV2", "StackedImages", "Stackedplt", "Stackedtorch",

           "plot_line", "bar_chart", "scatter_plot",]

RGB2BGR  = cv2.COLOR_RGB2BGR
BGR2RGB  = cv2.COLOR_BGR2RGB
BGR2GRAY = cv2.COLOR_BGR2GRAY
RGB2GRAY = cv2.COLOR_RGB2GRAY
GRAY2BGR = cv2.COLOR_GRAY2BGR
BGR2HSV  = cv2.COLOR_BGR2HSV
RGB2HSV  = cv2.COLOR_RGB2HSV
HSV2BGR  = cv2.COLOR_HSV2BGR
HSV2RGB  = cv2.COLOR_HSV2RGB

class ImagesReader(object):
    def __init__(self, filename, is_gray=False, flip=None):
        """
         Initialize ImagesReader object.

        :param filename: Path to the image file.
        :param is_gray: If True, read the image in grayscale.
        :param flip: 0: Flip vertically.
                    1: Flip horizontally.
                   -1: Flip both vertically and horizontally.
        """
        if not os.path.isfile(filename):
            raise ValueError(f"[ImagesReader]: The file {filename} does not exist.")
        if isinstance(filename, Path):
            filename = str(filename.resolve())
        self.filename = filename
        self.flags = cv2.IMREAD_COLOR
        self.is_gray = is_gray
        self.torgb = False
        if flip in [-1, 0, 1]:
            self.flip = flip
        else:
            raise ValueError("[ImagesReader]:The flip parameter should be -1, 0, or 1")

    def imread(self, mode=None):
        if mode is None or mode == "BGR":
            self.flags = cv2.IMREAD_COLOR
        elif self.is_gray or mode == "L":
            self.flags = cv2.IMREAD_GRAYSCALE
        elif mode == "RGB":
            self.torgb = True
        else:
            raise ValueError(f"[ImagesReader]: Unsupported mode: {mode}")

        self.image = cv2.imdecode(np.fromfile(self.filename, np.uint8), self.flags)

        if self.flip is not None:
            self.image = cv2.flip(self.image, self.flip)
        if self.torgb:
            self.image = cv2.cvtColor(self.image, cv2.COLOR_BGR2RGB)
        return self.image

    def display(self, winname):
        """Displays an image in the specified window."""
        _imshow = cv2.imshow  # copy to avoid recursion errors
        _imshow(winname.encode('unicode_escape').decode(), self.image)
        print(f"\033[94m{self.param()}")
        cv2.waitKey(0)

    def param(self, idx=None):
        properties = {}
        if isinstance(self.filename, str):
            properties['name'] = os.path.basename(self.filename)
            properties['format'] = imghdr.what(self.filename)
            properties['fsize'] = os.path.getsize(self.filename)
        properties["shape"] = self.image.shape
        properties["dtype"] = self.image.dtype
        properties["channels"] = self.image.shape[2] if len(self.image.shape) == 3 else 1
        properties["resolution"] = (self.image.shape[1], self.image.shape[0])
        return properties if idx is None else properties[idx]

def url2cvimg(url):
    if is_url(url):
        res = request.urlopen(url, timeout=3)
    else:
        raise ValueError("[url2imgcv2]:The current input parameter does not conform to the URL format")
    try:
        image = np.asarray(bytearray(res.read()), dtype="uint8")
        image = cv2.imdecode(image, cv2.IMREAD_UNCHANGED)
    except:
        print ('[url2imgcv2]: Load read - Image timeout!')
        image = []
    h, w, c = image.shape
    if c == 4:
        image = image[:,:,:3]
    return image

def imwrite(filename: str, img: np.ndarray, params=None):
    """Write the image to a file."""
    try:
        cv2.imencode(Path(filename).suffix, img, params)[1].tofile(filename)
        return True
    except Exception:
        return False

def imshowplt(image, title=None):
    """显示图像并添加标题"""
    import matplotlib
    matplotlib.use("TkAgg")
    cmap = None
    if is_pil:
        if image.mode == 'L' or (image.mode == 'LA' and image.getbands()[1] == 'L'):
            cmap = 'gray'
    else:
        if is_gray_image(image):
            cmap = 'gray'
    plt.imshow(image, cmap=cmap)
    plt.axis('off')
    if title:
        plt.title(title)
    plt.show()

def StackedImages(scale, imgArray):
    """
    :param scale:图片的规模,可使用小数,1为原始图像
    :param imgArray: 图像组成的列表，来表示行列排列
    :return: 生成的图像可按输入列表的行列排序展示,拼接展示图像
    """
    rows = len(imgArray)
    cols = len(imgArray[0])
    rowsAvailable = isinstance(imgArray[0], list)
    width = imgArray[0][0].shape[1]
    height = imgArray[0][0].shape[0]
    if rowsAvailable:
        for x in range(0, rows):
            for y in range(0, cols):
                if imgArray[x][y].shape[:2] == imgArray[0][0].shape [:2]:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (0, 0), None, scale, scale)
                else:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (imgArray[0][0].shape[1], imgArray[0][0].shape[0]), None, scale, scale)
                if len(imgArray[x][y].shape) == 2: imgArray[x][y]= cv2.cvtColor( imgArray[x][y], cv2.COLOR_GRAY2BGR)
        imageBlank = np.zeros((height, width, 3), np.uint8)
        hor = [imageBlank] * rows
        for x in range(0, rows):
            hor[x] = np.hstack(imgArray[x])
        ver = np.vstack(hor)
    else:
        for x in range(0, rows):
            if imgArray[x].shape[:2] == imgArray[0].shape[:2]:
                imgArray[x] = cv2.resize(imgArray[x], (0, 0), None, scale, scale)
            else:
                imgArray[x] = cv2.resize(imgArray[x], (imgArray[0].shape[1], imgArray[0].shape[0]), None,scale, scale)
            if len(imgArray[x].shape) == 2:
                imgArray[x] = cv2.cvtColor(imgArray[x], cv2.COLOR_GRAY2BGR)
        hor= np.hstack(imgArray)
        ver = hor
    return ver

def StackedCV2(scale, imgList, cols):
    """
    将多张图像堆叠在一起并在单个窗口中显示
    :param scale: 图像的缩放比例,大于1表示放大,小于1表示缩小
    :param imgList: 要堆叠的图像列表
    :param cols: 每行显示的图像数量
    :return: 堆叠后的图像
    """
    totalImages = len(imgList)
    rows = totalImages // cols if totalImages // cols * cols == totalImages else totalImages // cols + 1
    blankImages = cols * rows - totalImages

    width = imgList[0].shape[1]
    height = imgList[0].shape[0]
    imgBlank = np.zeros((height, width, 3), np.uint8)
    imgList.extend([imgBlank] * blankImages)
    for i in range(cols * rows):
        imgList[i] = cv2.resize(imgList[i], (0, 0), None, scale, scale)
        if len(imgList[i].shape) == 2:
            imgList[i] = cv2.cvtColor(imgList[i], cv2.COLOR_GRAY2BGR)
    hor = [imgBlank] * rows
    for y in range(rows):
        line = []
        for x in range(cols):
            line.append(imgList[y * cols + x])
        hor[y] = np.hstack(line)
    ver = np.vstack(hor)
    return ver

def Stackedplt(images, shape, spacing=0.1):
    """
    按照指定形状堆叠并显示多张图片——plt
    :param images: 包含多张图片数据的列表或数组。每张图片的形状应相同
    :param shape: 展示的图片的行列数,如(2, 3)
    :param spacing: 图片之间的间隔,默认为0.1
    :return:
    """
    rows, cols = shape
    image_height, image_width, _ = images[0].shape
    stacked_height = int(rows * image_height + (rows - 1) * image_height * spacing)
    stacked_width = int(cols * image_width + (cols - 1) * image_width * spacing)
    stacked_image = np.ones((stacked_height, stacked_width, 3), dtype=np.uint8) * 255
    for i, image in enumerate(images):
        row = i // cols
        col = i % cols

        x = int(col * (image_width + spacing * image_width))
        y = int(row * (image_height + spacing * image_height))

        stacked_image[y:y+image_height, x:x+image_width, :] = image

    plt.imshow(stacked_image)
    plt.axis('off')  # 关闭坐标轴显示
    plt.show()

def Stackedtorch(imgs, num_rows, num_cols, titles=None, scale=None, camp=None):
    """
    堆叠显示Tensor图像或PIL图像
    :param imgs: 包含图像的列表,可以包含Tensor图像或PIL图像
    :param num_rows: 图像的行数
    :param num_cols: 图像的列数
    :param titles: 图像的标题列表,默认为None
    :param scale: 图像的缩放比例,默认为None
    :param camp: 'gray'
    :return:
    example:
        imgs = [image1, image2, image3, image4]
        titles = ['Image 1', 'Image 2', 'Image 3','Image 4']
        axes = Stackedtorch(imgs, 2, 2, titles=titles, scale=5)
        plt.savefig('my_plot.jpg')
        plt.show()
    """
    if scale:
        figsize = (num_cols * scale, num_rows * scale)
        _, axes = plt.subplots(num_rows, num_cols, figsize=figsize)
    else:
        _, axes = plt.subplots(num_rows, num_cols)
    axes = axes.flatten()
    for i, (ax, img) in enumerate(zip(axes, imgs)):
        if torch.is_tensor(img):
            # Tensor Image
            ax.imshow(img.numpy(),cmap=camp)
        else:
            # PIL Image
            ax.imshow(img,cmap=camp)
        ax.axes.get_xaxis().set_visible(False)
        ax.axes.get_yaxis().set_visible(False)
        if titles:
            ax.set_title(titles[i])
    return axes

def plot_line(x, y, xlabel, ylabel, title):
    """
    绘制折线图
    :param x: x轴的数据列表
    :param y: y轴的数据列表
    :param xlabel: x轴的标签
    :param ylabel: y轴的标签
    :param title: 图形的标题
    :return:
    """
    plt.plot(x, y)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.show()

def bar_chart(x, y, xlabel, ylabel, title):
    """绘制柱状图"""
    plt.bar(x, y)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.show()

def scatter_plot(x, y, xlabel, ylabel, title):
    """绘制散点图"""
    plt.scatter(x, y)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.show()


if __name__=="__main__":
    x = [1, 2, 3, 4, 5]
    y = [2, 4, 6, 8, 10]
    xlabel = 'X'
    ylabel = 'Y'
    title = '折线图示例'
    plot_line(x, y, xlabel, ylabel, title)
    bar_chart(x, y, xlabel, ylabel, title)
    scatter_plot(x, y, xlabel, ylabel, title)



"""
论文原址： <https://arxiv.org/abs/1602.07360>
SQUEEZENET: ALEXNET-LEVEL ACCURACY WITH
50X FEWER PARAMETERS AND <0.5MB MODEL SIZE
"""
import torch
import torch.nn as nn
import torch.nn.init as init

__all__ = ['SqueezeNet', 'squeezenet1_0', 'squeezenet1_1']


class Fire(nn.Module):
    def __init__(self, inplanes, squeeze_planes, expand1x1_planes, expand3x3_planes):
        super(Fire, self).__init__()
        self.inplanes = inplanes
        self.squeeze = nn.Conv2d(inplanes, squeeze_planes, kernel_size=1)
        self.squeeze_activation = nn.ReLU(inplace=True)
        self.expand1x1 = nn.Conv2d(squeeze_planes, expand1x1_planes,
                                   kernel_size=1)
        self.expand1x1_activation = nn.ReLU(inplace=True)
        self.expand3x3 = nn.Conv2d(squeeze_planes, expand3x3_planes,
                                   kernel_size=3, padding=1)
        self.expand3x3_activation = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.squeeze_activation(self.squeeze(x))
        return torch.cat([
            self.expand1x1_activation(self.expand1x1(x)),
            self.expand3x3_activation(self.expand3x3(x))
        ], dim=1)


class SqueezeNet(nn.Module):
    def __init__(self, version='1_0', num_classes=1000):
        super(SqueezeNet, self).__init__()
        self.num_classes = num_classes
        if version == '1_0':
            self.features = nn.Sequential(
                nn.Conv2d(3, 96, kernel_size=7, stride=2),
                nn.ReLU(inplace=True),
                nn.MaxPool2d(kernel_size=3, stride=2, ceil_mode=True),
                Fire(96, 16, 64, 64),
                Fire(128, 16, 64, 64),
                Fire(128, 32, 128, 128),
                nn.MaxPool2d(kernel_size=3, stride=2, ceil_mode=True),
                Fire(256, 32, 128, 128),
                Fire(256, 48, 192, 192),
                Fire(384, 48, 192, 192),
                Fire(384, 64, 256, 256),
                nn.MaxPool2d(kernel_size=3, stride=2, ceil_mode=True),
                Fire(512, 64, 256, 256),
            )
        elif version == '1_1':
            self.features = nn.Sequential(
                nn.Conv2d(3, 64, kernel_size=3, stride=2),
                nn.ReLU(inplace=True),
                nn.MaxPool2d(kernel_size=3, stride=2, ceil_mode=True),
                Fire(64, 16, 64, 64),
                Fire(128, 16, 64, 64),
                nn.MaxPool2d(kernel_size=3, stride=2, ceil_mode=True),
                Fire(128, 32, 128, 128),
                Fire(256, 32, 128, 128),
                nn.MaxPool2d(kernel_size=3, stride=2, ceil_mode=True),
                Fire(256, 48, 192, 192),
                Fire(384, 48, 192, 192),
                Fire(384, 64, 256, 256),
                Fire(512, 64, 256, 256),
            )
        else:
            raise ValueError("Unsupported SqueezeNet version {version}:"
                             "1_0 or 1_1 expected".format(version=version))

        # Final convolution is initialized differently from the rest
        final_conv = nn.Conv2d(512, self.num_classes, kernel_size=1)
        self.classifier = nn.Sequential(
            nn.Dropout(p=0.5),
            final_conv,
            nn.ReLU(inplace=True),
            nn.AdaptiveAvgPool2d((1, 1))
        )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m is final_conv:
                    init.normal_(m.weight, mean=0.0, std=0.01)
                else:
                    init.kaiming_uniform_(m.weight)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.features(x)
        x = self.classifier(x)
        return torch.flatten(x, 1)


def squeezenet1_0(num_classes):
    r"""SqueezeNet model architecture from the `"SqueezeNet: AlexNet-level
    accuracy with 50x fewer parameters and <0.5MB model size"
    <https://arxiv.org/abs/1602.07360>`_ paper.
    """
    return SqueezeNet('1_0', num_classes)


def squeezenet1_1(num_classes):
    r"""SqueezeNet 1.1 model from the `official SqueezeNet repo
    <https://github.com/DeepScale/SqueezeNet/tree/master/SqueezeNet_v1.1>`_.
    SqueezeNet 1.1 has 2.4x less computation and slightly fewer parameters
    than SqueezeNet 1.0, without sacrificing accuracy.
    """
    return SqueezeNet('1_1', num_classes)


if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = squeezenet1_0(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # Total params: 737,476
    #               724,548


import datetime

def printlog(info):
    nowtime = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    print("\n"+"=========="*8 + "%s"%nowtime)
    print(info+'...\n\n')

def colorful(obj, color="red", display_type="plain"):
    # 彩色输出格式：
    # 设置颜色开始 ：\033[显示方式;前景色;背景色m
    # 前景色            背景色           颜色
    # ---------------------------------------
    # 30                40              黑色
    # 31                41              红色
    # 32                42              绿色
    # 33                43              黃色
    # 34                44              蓝色
    # 35                45              紫红色
    # 36                46              青蓝色
    # 37                47              白色
    # 显示方式           意义
    # -------------------------
    # 0                终端默认设置
    # 1                高亮显示
    # 4                使用下划线
    # 5                闪烁
    # 7                反白显示
    # 8                不可见
    color_dict = {"black":"30", "red":"31", "green":"32", "yellow":"33",
                  "blue":"34", "purple":"35","cyan":"36",  "white":"37"}
    display_type_dict = {"plain":"0","highlight":"1","underline":"4",
                         "shine":"5","inverse":"7","invisible":"8"}
    s = str(obj)
    color_code = color_dict.get(color, "")
    display = display_type_dict.get(display_type, "")
    out = '\033[{};{}m'.format(display, color_code) + s + '\033[0m'
    return out


def colorstr(*input):
    # https://en.wikipedia.org/wiki/ANSI_escape_code, i.e.  colorstr('blue', 'hello world')
    *args, string = input if len(input) > 1 else ('blue', 'bold', input[0])
    colors = {
        'black': '\033[30m',  # basic colors
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'magenta': '\033[35m',
        'cyan': '\033[36m',
        'white': '\033[37m',
        'bright_black': '\033[90m',  # bright colors
        'bright_red': '\033[91m',
        'bright_green': '\033[92m',
        'bright_yellow': '\033[93m',
        'bright_blue': '\033[94m',
        'bright_magenta': '\033[95m',
        'bright_cyan': '\033[96m',
        'bright_white': '\033[97m',
        'end': '\033[0m',  # misc
        'bold': '\033[1m',
        'underline': '\033[4m'}
    return ''.join(colors[x] for x in args) + f'{string}' + colors['end']



if __name__=="__main__":
    text = "Hello, Colorful World!"
    printlog("Starting the main process.")

    colors = ["red", "green", "blue", "yellow", "purple", "cyan", "white"]
    display_types = ["plain", "highlight", "underline", "shine", "inverse", "invisible"]

    for color in colors:
        for display_type in display_types:
            colored_text = colorful(text, color=color, display_type=display_type)
            print(f"Color: {color}, Display Type: {display_type} -> {colored_text}")





"""
Haar Wavelet-based Downsampling (HWD)

Original address of the paper: https://www.sciencedirect.com/science/article/abs/pii/S0031320323005174
Code reference: https://github.com/apple1986/HWD/tree/main
Time: 2024-01-22 17:11

Blog records: https://blog.csdn.net/m0_62919535/article/details/135758371
I conducted comparative experiments using my own dataset.
"""
import torch
import torch.nn as nn
from pytorch_wavelets import DWTForward

class HWDownsampling(nn.Module):
    def __init__(self, in_channel, out_channel):
        super(HWDownsampling, self).__init__()
        self.wt = DWTForward(J=1, wave='haar', mode='zero')
        self.conv_bn_relu = nn.Sequential(
            nn.Conv2d(in_channel * 4, out_channel, kernel_size=1, stride=1),
            nn.BatchNorm2d(out_channel),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        yL, yH = self.wt(x)
        y_HL = yH[0][:, :, 0, ::]
        y_LH = yH[0][:, :, 1, ::]
        y_HH = yH[0][:, :, 2, ::]
        x = torch.cat([yL, y_HL, y_LH, y_HH], dim=1)
        x = self.conv_bn_relu(x)
        return x


if __name__ == '__main__':
    downsampling_layer = HWDownsampling(3, 64)
    input_data = torch.rand((1, 3, 64, 64))
    output_data = downsampling_layer(input_data)
    print("Input shape:", input_data.shape)
    print("Output shape:", output_data.shape)


"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is based on HSV for color detection.
"""
import cv2
import numpy as np
import logging
from pyzjr.videos.utils import VideoCap
import pyzjr.Z as Z

def empty(a):
    pass

class ColorFind():
    def __init__(self, trackBar=False, name="Bars"):
        self.trackBar = trackBar
        self.name = name
        if self.trackBar:
            self.initTrackbars()

    def initTrackbars(self):
        """
        :return:初始化轨迹栏
        """
        cv2.namedWindow(self.name)
        cv2.resizeWindow(self.name, 640, 240)
        cv2.createTrackbar("Hue Min", self.name, 0, 179, empty)
        cv2.createTrackbar("Hue Max", self.name, 179, 179, empty)
        cv2.createTrackbar("Sat Min", self.name, 0, 255, empty)
        cv2.createTrackbar("Sat Max", self.name, 255, 255, empty)
        cv2.createTrackbar("Val Min", self.name, 0, 255, empty)
        cv2.createTrackbar("Val Max", self.name, 255, 255, empty)

    def getTrackbarValues(self):
        """
        Gets the trackbar values in runtime
        :return: hsv values from the trackbar window
        """
        hmin = cv2.getTrackbarPos("Hue Min", self.name)
        smin = cv2.getTrackbarPos("Sat Min", self.name)
        vmin = cv2.getTrackbarPos("Val Min", self.name)
        hmax = cv2.getTrackbarPos("Hue Max", self.name)
        smax = cv2.getTrackbarPos("Sat Max", self.name)
        vmax = cv2.getTrackbarPos("Val Max", self.name)
        HsvVals=[[hmin, smin, vmin],[hmax, smax, vmax]]

        return HsvVals

    def protect_region(self, mask ,threshold=None):
        """
        * 用于保护掩膜图的部分区域
        :param mask: 掩膜图
        :param threshold: 如果为None,则为不保护，如果是长为4的列表，则进行特定区域的保护
        :return: 返回进行保护区域的掩膜图

        example:    [0, img.shape[1], 0, img.shape[0]]为全保护状态，
                    x_start可以保护大于x的部分
                    x_end可以保护小于x的部分
                    y_start可以保护图像下方的部分
                    y_end可以保护图像上方的部分
        """
        if threshold == None:
            return mask
        else:
            x_start, x_end, y_start, y_end = threshold[:4]
            mask[y_start:y_end, x_start:x_end] = 0
            return mask

    def MaskZone(self, img, HsvVals):
        """
        * 生成掩膜图以及融合图像
        :param img: 输入图像
        :param HsvVals: 可以通过getTrackbarValues获得,也可调取Z.HSV的值
        :return: 返回融合图、掩膜图、HSV图
        """
        imgHSV = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
        lower = np.array(HsvVals[0])
        upper = np.array(HsvVals[1])
        mask = cv2.inRange(imgHSV, lower, upper)
        imgResult = cv2.bitwise_and(img, img, mask=mask)
        return imgResult, mask

    def update(self, img, myColor=None):
        """
        :param img: 需要在其中找到颜色的图像
        :param myColor: hsv上下限列表
        :return: mask带有检测到颜色的白色区域的roi图像
                 imgColor彩色图像仅显示检测到的区域
        """
        imgColor = [],
        mask = []
        if self.trackBar:
            myColor = self.getTrackbarValues()

        if isinstance(myColor, str):
            myColor = self.getColorHSV(myColor)

        if myColor is not None:
            imgColor, mask = self.MaskZone(img,myColor)
        return imgColor, mask

    def getColorHSV(self, myColor):
        if myColor == 'red':
            output = [[146, 141, 77], [179, 255, 255]]
        elif myColor == 'green':
            output = [[44, 79, 111], [79, 255, 255]]
        elif myColor == 'blue':
            output = [[103, 68, 130], [128, 255, 255]]
        else:
            output = None
            logging.warning("Color Not Defined")
            logging.warning("Available colors: red, green, blue ")

        return output


if __name__=="__main__":
    def DetectImg(img, name="DetectImg", ConsoleOut=True, threshold=None, scale=1.0):
        """
        * 轨迹栏检测图片,此函数仅仅作为使用示例
        :param img: 图片
        :param name: 轨迹栏名
        :param ConsoleOut: 用于是否控制台打印HsvVals的值
        :param threshold: 阈值，用于保护图片的区域
        :param scale: 规模大小
        :return:
        """
        ColF = ColorFind(True, name)
        while True:
            imgHSV = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
            HsvVals = ColF.getTrackbarValues()
            if ConsoleOut:
                print(HsvVals)
            imgResult, mask = ColF.update(img,HsvVals)
            pro_mask = ColF.protect_region(mask, threshold)
            imgStack = StackedImages(scale, ([img,imgHSV],[pro_mask,imgResult]))
            cv2.imshow("Stacked Images", imgStack)
            k = cv2.waitKey(1)
            if k == Z.Esc:
                break

    def DetectVideo(name="DetectVideo",mode=0,myColor=None,scale=1.0):
        """
        * 轨迹栏检测摄像头,此函数仅仅作为使用示例
        :param mode: 检测模式,默认本地摄像头,可传入video路径
        :param name: 轨迹栏名
        :param myColor: getColorHSV返回的一些测好的Hsv值
        :param scale: 规模大小
        :return:
        """
        if myColor:
            Cf = False
        else:
            Cf = True
        Vcap = VideoCap()
        Vcap.CapInit(mode=mode)
        ColF = ColorFind(Cf, name)
        while True:
            img = Vcap.read()
            imgColor, mask = ColF.update(img, myColor)
            stackedimg=StackedImages(scale,[img,imgColor])
            cv2.imshow(name, stackedimg)
            if cv2.waitKey(1) & 0xFF == Z.Esc:
                break

    def StackedImages(scale,imgArray):
        """
        :param scale:图片的规模,可使用小数,1为原始图像
        :param imgArray: 图像组成的列表，来表示行列排列
        :return: 生成的图像可按输入列表的行列排序展示,拼接展示图像
        """
        rows = len(imgArray)
        cols = len(imgArray[0])
        rowsAvailable = isinstance(imgArray[0], list)
        width = imgArray[0][0].shape[1]
        height = imgArray[0][0].shape[0]
        if rowsAvailable:
            for x in range(0, rows):
                for y in range(0, cols):
                    if imgArray[x][y].shape[:2] == imgArray[0][0].shape [:2]:
                        imgArray[x][y] = cv2.resize(imgArray[x][y], (0, 0), None, scale, scale)
                    else:
                        imgArray[x][y] = cv2.resize(imgArray[x][y], (imgArray[0][0].shape[1], imgArray[0][0].shape[0]), None, scale, scale)
                    if len(imgArray[x][y].shape) == 2: imgArray[x][y]= cv2.cvtColor( imgArray[x][y], cv2.COLOR_GRAY2BGR)
            imageBlank = np.zeros((height, width, 3), np.uint8)
            hor = [imageBlank]*rows
            for x in range(0, rows):
                hor[x] = np.hstack(imgArray[x])
            ver = np.vstack(hor)
        else:
            for x in range(0, rows):
                if imgArray[x].shape[:2] == imgArray[0].shape[:2]:
                    imgArray[x] = cv2.resize(imgArray[x], (0, 0), None, scale, scale)
                else:
                    imgArray[x] = cv2.resize(imgArray[x], (imgArray[0].shape[1], imgArray[0].shape[0]), None,scale, scale)
                if len(imgArray[x].shape) == 2:
                    imgArray[x] = cv2.cvtColor(imgArray[x], cv2.COLOR_GRAY2BGR)
            hor= np.hstack(imgArray)
            ver = hor
        return ver
    DetectVideo(myColor="red")




"""
Copyright (c) 2024, Auorui.
All rights reserved.

RegNet: <https://arxiv.org/pdf/2003.13678.pdf>
The source code comes from: https://github.com/d-li14/regnet.pytorch/blob/master/regnet.py
"""
import torch
import torch.nn as nn


__all__ = ['regnetx_002', 'regnetx_004', 'regnetx_006', 'regnetx_008', 'regnetx_016', 'regnetx_032',
           'regnetx_040', 'regnetx_064', 'regnetx_080', 'regnetx_120', 'regnetx_160', 'regnetx_320']

def conv3x3(in_channels, out_channels, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding:捕捉局部特征和空间相关性，学习更复杂的特征和抽象表示"""
    return nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)

def conv1x1(in_channels, out_channels, stride=1):
    """1x1 convolution:实现降维或升维，调整通道数和执行通道间的线性变换"""
    return nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)


class Bottleneck(nn.Module):
    expansion = 1
    def __init__(self, inplanes, planes, stride=1, downsample=None, group_width=1, dilation=1):
        super(Bottleneck, self).__init__()
        width = planes * self.expansion
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = nn.BatchNorm2d(width)
        self.conv2 = conv3x3(width, width, stride, width // min(width, group_width), dilation)
        self.bn2 = nn.BatchNorm2d(width)
        self.conv3 = conv1x1(width, planes)
        self.bn3 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class RegNet(nn.Module):
    def __init__(self, block, layers, widths, num_classes=1000, group_width=1):
        super(RegNet, self).__init__()
        self.inplanes = 32
        self.dilation = 1
        replace_stride_with_dilation = [False, False, False, False]
        self.group_width = group_width
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=3, stride=2, padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.layer1 = self._make_layer(block, widths[0], layers[0], stride=2,
                                       dilate=replace_stride_with_dilation[0])
        self.layer2 = self._make_layer(block, widths[1], layers[1], stride=2,
                                       dilate=replace_stride_with_dilation[1])
        self.layer3 = self._make_layer(block, widths[2], layers[2], stride=2,
                                       dilate=replace_stride_with_dilation[2])
        self.layer4 = self._make_layer(block, widths[3], layers[3], stride=2,
                                       dilate=replace_stride_with_dilation[3])
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(widths[-1] * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, Bottleneck):
                nn.init.constant_(m.bn3.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes, stride),
                nn.BatchNorm2d(planes),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.group_width,
                            previous_dilation))
        self.inplanes = planes
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, group_width=self.group_width,
                                dilation=self.dilation))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x


def regnetx_002(**kwargs):
    return RegNet(Bottleneck, [1, 1, 4, 7], [24, 56, 152, 368], group_width=8, **kwargs)


def regnetx_004(**kwargs):
    return RegNet(Bottleneck, [1, 2, 7, 12], [32, 64, 160, 384], group_width=16, **kwargs)


def regnetx_006(**kwargs):
    return RegNet(Bottleneck, [1, 3, 5, 7], [48, 96, 240, 528], group_width=24, **kwargs)


def regnetx_008(**kwargs):
    return RegNet(Bottleneck, [1, 3, 7, 5], [64, 128, 288, 672], group_width=16, **kwargs)


def regnetx_016(**kwargs):
    return RegNet(Bottleneck, [2, 4, 10, 2], [72, 168, 408, 912], group_width=24, **kwargs)


def regnetx_032(**kwargs):
    return RegNet(Bottleneck, [2, 6, 15, 2], [96, 192, 432, 1008], group_width=48, **kwargs)


def regnetx_040(**kwargs):
    return RegNet(Bottleneck, [2, 5, 14, 2], [80, 240, 560, 1360], group_width=40, **kwargs)


def regnetx_064(**kwargs):
    return RegNet(Bottleneck, [2, 4, 10, 1], [168, 392, 784, 1624], group_width=56, **kwargs)


def regnetx_080(**kwargs):
    return RegNet(Bottleneck, [2, 5, 15, 1], [80, 240, 720, 1920], group_width=120, **kwargs)


def regnetx_120(**kwargs):
    return RegNet(Bottleneck, [2, 5, 11, 1], [224, 448, 896, 2240], group_width=112, **kwargs)


def regnetx_160(**kwargs):
    return RegNet(Bottleneck, [2, 6, 13, 1], [256, 512, 896, 2048], group_width=128, **kwargs)


def regnetx_320(**kwargs):
    return RegNet(Bottleneck, [2, 7, 13, 1], [336, 672, 1344, 2520], group_width=168, **kwargs)


if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = regnetx_064(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # 320 Total params: 105,300,644   Estimated Total Size (MB): 1289.42
    # 120 Total params: 43,874,020    Estimated Total Size (MB): 686.45
    # 080 Total params: 37,659,332    Estimated Total Size (MB): 488.76
    # 064 Total params: 24,590,756    Estimated Total Size (MB): 490.65


import math
import numpy as np

def Advancedlist(imglist,mode='T',reverse=False):
    """对列表的高级排列"""
    if mode=="T":
        """平铺
        lists = [[1, 2, 3], [4, 5, 6], [7, 8, 9]] ——> [1, 2, 3, 4, 5, 6, 7, 8, 9]
        """
        tiled_list = [item for sublist in imglist for item in sublist]
        return tiled_list
    elif mode=="F":
        """首元素
        lists = [[1, 2, 3], [4, 5, 6], [7, 8, 9]] ——> [1, 4, 7, 2, 5, 8, 3, 6, 9]
        """
        sorted_list_by_first_element = [sublist[i] for i in range(len(imglist[0])) for sublist in imglist]
        return sorted_list_by_first_element
    elif mode=="X":
        """按照x排序,默认是从小到大
        lists = [(1, 4), (3, 8), (5, 2)] ——> [(1, 4), (3, 8), (5, 2)]
        """
        sorted_by_x = sorted(imglist, key=lambda item: item[0], reverse=reverse)
        return sorted_by_x
    elif mode=="Y":
        """按照y排序,默认是从小到大
        lists = [(1, 4), (3, 8), (5, 2)] ——> [(5, 2), (1, 4), (3, 8)]
        """
        sorted_by_y = sorted(imglist, key=lambda item: item[1], reverse=reverse)
        return sorted_by_y

class GraphPoint:
    def __init__(self, x, y, z=None):
        self.x = x
        self.y = y
        self.z = z

    def __str__(self):
        if self.z is not None:
            return f"({self.x}, {self.y}, {self.z})"
        else:
            return f"({self.x}, {self.y})"

    def __sub__(self, other: 'GraphPoint') -> 'GraphPoint':
        return GraphPoint(self.x - other.x, self.y - other.y, self.z - other.z if self.z is not None and other.z is not None else None)

    def __mul__(self, scalar: float):
        return GraphPoint(self.x * scalar, self.y * scalar, self.z * scalar if self.z is not None else None)

    def __rmul__(self, scalar: float):
        return self.__mul__(scalar)

    def __truediv__(self, scalar: float):
        return GraphPoint(self.x / scalar, self.y / scalar, self.z / scalar if self.z is not None else None)

    def as_tuple(self):
        """
        GraphPoint转为tuple类型
        >>> p3d = GraphPoint(2, 3, 4)
        >>> p2d = GraphPoint(5, 6)
        >>> tuple_3d = p3d.as_tuple()
        >>> # (2, 3, 4)
        >>> tuple_2d = p2d.as_tuple()
        >>> # (5, 6)
        """
        if self.z is not None:
            return (self.x, self.y, self.z)
        else:
            return (self.x, self.y)

    def as_numpy(self):
        """
        GraphPoint转为numpy类型
        >>> p3d = GraphPoint(2, 3, 4)
        >>> np_array = p3d.as_numpy()
        >>> print(np_array)
        >>> print(type(np_array))
        >>> # [2 3 4]
        >>> # <class 'numpy.ndarray'>
        """
        if self.z is not None:
            return np.array([self.x, self.y, self.z])
        else:
            return np.array([self.x, self.y])

    @classmethod
    def from_numpy(cls, np_array):
        """
        numpy转为GraphPoint类型
        >>> new_p3d = GraphPoint.from_numpy(np_array)
        >>> print(new_p3d)
        >>> print(type(new_p3d))
        >>> # Point(2, 3, 4)
        >>> # <class '__main__.GraphPoint'>
        """
        if len(np_array) == 2:
            return cls(np_array[0], np_array[1])
        elif len(np_array) == 3:
            return cls(np_array[0], np_array[1], np_array[2])
        else:
            raise ValueError("Invalid array shape for conversion.")

    def rotate(self, angle_degrees: float):
        """
        将点逆时针旋转45度
        >>> p = GraphPoint(1, 0)  # 创建一个点 (1, 0)
        >>> rotated_point = p.rotate(90)
        >>> # 输出: (6.123233995736766e-17, 1.0)
        将点逆时针旋转45度
        >>> rotated_point = p.rotate(45)
        >>> # 输出: (0.7071067811865476, 0.7071067811865476)
        """
        angle_radians = math.radians(angle_degrees)
        cos_angle = math.cos(angle_radians)
        sin_angle = math.sin(angle_radians)
        x_rotated = self.x * cos_angle - self.y * sin_angle
        y_rotated = self.x * sin_angle + self.y * cos_angle
        return GraphPoint(x_rotated, y_rotated, self.z)

    def distance(self, other_point2: "GraphPoint"):
        """
        计算两点间间距
        >>> p = GraphPoint(1, 0)
        >>> p2 = GraphPoint(2, 2)
        >>> distance = p.distance(p2)
        >>> # 2.23606797749979
        """
        dx = other_point2.x - self.x
        dy = other_point2.y - self.y
        if self.z is not None and other_point2.z is not None:
            dz = other_point2.z - self.z
            return math.sqrt(dx**2 + dy**2 + dz**2)
        else:
            return math.sqrt(dx**2 + dy**2)

    def move(self, dx, dy, dz=None):
        """
        在x轴上移动dx个单位，y轴上移动dy个单位，z轴上移动dz个单位
        >>> p = GraphPoint(1, 0, 2)
        >>> p.move(2, 3, 1)  #
        >>> # Point(3, 3, 3)
        """
        self.x += dx
        self.y += dy
        if dz is not None and self.z is not None:
            self.z += dz

    def rotate_by(self, angle_degrees, origin=None):
        """
        绕着某点进行旋转
        >>> p = GraphPoint(1, 0)
        >>> origin = GraphPoint(0, 0)
        >>> p.rotate_by(90, origin)
        >>> # Point(6.123233995736766e-17, 1.0)
        """
        if origin is None:
            origin = GraphPoint(0, 0)

        angle_radians = math.radians(angle_degrees)
        cos_angle = math.cos(angle_radians)
        sin_angle = math.sin(angle_radians)

        translated_x = self.x - origin.x
        translated_y = self.y - origin.y

        rotated_x = translated_x * cos_angle - translated_y * sin_angle
        rotated_y = translated_x * sin_angle + translated_y * cos_angle

        self.x = rotated_x + origin.x
        self.y = rotated_y + origin.y


    def scale(self, factor):
        """
        缩放二维或三维点的坐标
        >>> p3d = GraphPoint(2, 3, 4)
        >>> p3d.scale(2)
        >>> # (4, 6, 8)
        """
        self.x *= factor
        self.y *= factor
        if self.z is not None:
            self.z *= factor

    def midpoint_to(self, other_point:"GraphPoint"):
        """
        中点计算
        >>> p1 = GraphPoint(1, 2, 3)
        >>> p2 = GraphPoint(4, 5, 6)
        >>> mid_point = p1.midpoint_to(p2)
        >>> # (2.5, 3.5, 4.5)
        """
        mx = (self.x + other_point.x) / 2
        my = (self.y + other_point.y) / 2
        if self.z is not None and other_point.z is not None:
            mz = (self.z + other_point.z) / 2
            return GraphPoint(mx, my, mz)
        else:
            return GraphPoint(mx, my)

    def reflect(self, origin=None):
        """
        反射对称点,默认是原点对称
        >>> p = GraphPoint(2, 3, 4)
        >>> reflected_point = p.reflect()
        >>> # (-2, -3, -4)
        """
        if origin is None:
            origin = GraphPoint(0, 0, z=0)

        rx = 2 * origin.x - self.x
        ry = 2 * origin.y - self.y
        if self.z is not None and origin.z is not None:
            rz = 2 * origin.z - self.z
            return GraphPoint(rx, ry, rz)
        else:
            return GraphPoint(rx, ry)

    def angle_between(self, other_point:"GraphPoint"):
        """
        以度为单位,两点间的角度
        >>> p1 = GraphPoint(1, 0)
        >>> p2 = GraphPoint(0, 1)
        >>> angle_degrees = p1.angle_between(p2)
        >>> # 135.0
        """
        if self.z is not None or other_point.z is not None:
            raise ValueError("Angle between is not supported for 3D points.")
        dx = other_point.x - self.x
        dy = other_point.y - self.y
        angle_rad = math.atan2(dy, dx)
        return math.degrees(angle_rad)

    def polar_coordinates(self):
        """
        笛卡尔转为极坐标
        >>> p2d = GraphPoint(3, 4)  # 创建一个二维点 (3, 4)
        >>> r, theta = p2d.polar_coordinates()
        >>> print(f"r = {r}, theta = {theta} degrees")
        >>> # r = 5.0, theta = 53.13010235415598 degrees
        """
        if self.z is not None:
            raise ValueError("Polar coordinates are not supported for 3D points.")

        r = math.sqrt(self.x ** 2 + self.y ** 2)
        theta = math.degrees(math.atan2(self.y, self.x))
        return r, theta

    def set_polar_coordinates(self, r, theta, origin=None):
        """
        将极坐标系转为笛卡尔坐标系
        >>> p2d = GraphPoint(0, 0)
        >>> p2d.set_polar_coordinates(5, 45)
        >>> # (3.5355339059327378, 3.5355339059327378)
        """
        if origin is None:
            origin = GraphPoint(0, 0)

        theta_radians = math.radians(theta)
        self.x = origin.x + r * math.cos(theta_radians)
        self.y = origin.y + r * math.sin(theta_radians)

    def slope(self, other_point: "GraphPoint"):
        """
        斜率
        >>> p1 = GraphPoint(1, 2)
        >>> p2 = GraphPoint(3, 4)

        >>> slope = p1.slope(p2)
        >>> # 1.0
        """
        dy = other_point.y - self.y
        dx = other_point.x - self.x
        if dx == 0:
            return float('inf')
        return dy / dx

    def is_collinear(self, a:"GraphPoint", b:"GraphPoint"):
        """
        两条线是否共线
        >>> p1 = GraphPoint(1, 1)
        >>> p2 = GraphPoint(2, 2)
        >>> p3 = GraphPoint(3, 3)
        >>> collinear = p1.is_collinear(p2, p3)
        >>> # True
        """
        return (b.y - a.y) * (b.x - self.x) == (b.y - self.y) * (b.x - a.x)

    def distance_to_line(self, line_point1:"GraphPoint", line_point2:"GraphPoint"):
        """
        点到线段的最短距离
        1,计算线段的方向向量 (dx, dy):
            dx = x2 - x1
            dy = y2 - y1
        2,计算点 (x0, y0) 到线段起点 (x1, y1) 的向量 (dpx, dpy)：
            dpx = x0 - x1
            dpy = y0 - y1
        3,计算点 (x0, y0) 到线段的垂直距离 dist，使用向量叉积：
            dist = |dpx * dy - dpy * dx| / sqrt(dx^2 + dy^2)
        """
        num = abs((line_point2.y - line_point1.y) * self.x - (line_point2.x - line_point1.x) * self.y + line_point2.x * line_point1.y - line_point2.y * line_point1.x)
        den = ((line_point2.y - line_point1.y)**2 + (line_point2.x - line_point1.x)**2)**0.5
        return num / den

    def point_to(self, other_point: "GraphPoint"):
        """
        点指向目标点的方向
        >>> point1 = GraphPoint(1, 2)
        >>> point2 = GraphPoint(4, 6)
        >>> angle = point1.point_to(point2)
        >>> # 53.13010235415598
        """
        angle = math.atan2(other_point.y - self.y, other_point.x - self.x)
        bearing = math.degrees(angle) % 360
        return bearing

    def distance_to_circle(self, center:"GraphPoint", radius):
        """
        计算点到圆的距离，如果点在圆内，则返回0表示在圆内
        >>> p = GraphPoint(5, 4)
        >>> center = GraphPoint(0, 0)
        >>> radius = 5
        >>> distance = p.distance_to_circle(center, radius)
        >>> # 1.4031242374328485
        """
        distance_to_center = self.distance(center)
        return max(0, distance_to_center - radius)

    def distance_to_pline(self, polyline):
        """
        计算点到多段线的最短距离
        >>> p = GraphPoint(3, 3)
        >>> polyline = [GraphPoint(3, 2), GraphPoint(2, 7), GraphPoint(1, 5)]
        >>> distance = p.distance_to_pline(polyline)
        >>> # 0.19611613513818404
        """
        min_distance = float("inf")
        for i in range(len(polyline) - 1):
            distance = self.distance_to_line(polyline[i], polyline[i + 1])
            min_distance = min(min_distance, distance)

        return min_distance

    def interpolate(self, point1, point2, ratio):
        """
        在两个点之间按照给定的比例插值出一个新的点
        >>> point = GraphPoint(0, 0, 0)
        >>> point1 = GraphPoint(1, 2, 3)
        >>> point2 = GraphPoint(4, 5, 6)
        >>> ratio = 0.5
        >>> interpolated_point = point.interpolate(point1, point2, ratio)
        >>> # (2.5, 3.5, 4.5)
        """
        x = point1.x + ratio * (point2.x - point1.x)
        y = point1.y + ratio * (point2.y - point1.y)
        if self.z is not None and point1.z is not None and point2.z is not None:
            z = point1.z + ratio * (point2.z - point1.z)
            return GraphPoint(x, y, z)
        return GraphPoint(x, y)

    def triangle_area(self, point1, point2):
        """
        >>> point = GraphPoint(0, 0)
        >>> point1 = GraphPoint(1, 2)
        >>> point2 = GraphPoint(4, 5)
        >>> area = point.triangle_area(point1, point2)
        >>> # 1.5
        """
        area = 0.5 * abs((point1.x - self.x) * (point2.y - self.y) - (point1.y - self.y) * (point2.x - self.x))
        return area

    def is_inside_polygon(self, polygon):
        """
        射线法检查点是否在多边形内部
        >>> polygon = [GraphPoint(1, 1),GraphPoint(2, 2),GraphPoint(4, 3),GraphPoint(4, 1)]
        >>> test_point = GraphPoint(3, 2)

        >>> is_inside = test_point.is_inside_polygon(polygon)
        >>> # True
        """
        num_vertices = len(polygon)
        j = num_vertices - 1
        inside = False

        for i in range(num_vertices):
            if ((polygon[i].y > self.y) != (polygon[j].y > self.y)) and \
                    (self.x < (polygon[j].x - polygon[i].x) * (self.y - polygon[i].y) / (polygon[j].y - polygon[i].y) + polygon[i].x):
                inside = not inside
            j = i
        return inside

class GraphLine():
    def __init__(self, point1: GraphPoint, point2: GraphPoint):
        if point1 == point2:
            raise ValueError("The two points must be distinct.")
        self.point1 = point1
        self.point2 = point2

    def __repr__(self):
        return f"({self.point1}, {self.point2})"

    def length(self):
        """计算线段长度"""
        return self.point1.distance(self.point2)

    def slope(self):
        """计算线段斜率"""
        return (self.point2.y - self.point1.y) / (self.point2.x - self.point1.x)

    def intercept(self):
        """计算线段于y轴的截距"""
        return self.point1.y - self.slope() * self.point1.x

    def midpoint(self):
        """计算线段的中点"""
        return self.point1.midpoint_to(self.point2)

    def contains_point(self, point):
        """判断点是否在线段上"""
        return point.is_collinear(self.point1, self.point2) and self.point1.distance(point) + self.point2.distance(point) == self.length()

    def perpendicular_bisector(self):
        """构造线段的垂线"""
        midpoint = self.midpoint()
        if self.point2.x == self.point1.x:
            return GraphLine(midpoint, GraphPoint(midpoint.x + 1, midpoint.y))
        slope = -1 / self.slope()
        intercept = midpoint.y - slope * midpoint.x
        return GraphLine(midpoint, GraphPoint(midpoint.x + 1, slope * (midpoint.x + 1) + intercept))

    def angle_between(self, other_line):
        """计算两线段之间的角度"""
        tan_angle = abs((self.slope() - other_line.slope()) / (1 + self.slope() * other_line.slope()))
        return math.degrees(math.atan(tan_angle))

    def point_at_distance(self, point, distance):
        """返回线段上距离给定点指定距离的点"""
        if not self.contains_point(point):
            raise ValueError("The given point must be on the line.")

        line_length = self.length()
        ratio = distance / line_length

        x = point.x + ratio * (self.point2.x - self.point1.x)
        y = point.y + ratio * (self.point2.y - self.point1.y)

        return GraphPoint(x, y)

    def intersection(self, other_line):
        """返回两条直线的交点坐标"""
        if self.slope() == other_line.slope():
            return None
        x = (other_line.intercept() - self.intercept()) / (self.slope() - other_line.slope())
        y = self.slope() * x + self.intercept()
        return GraphPoint(x, y)

    def parallel_to(self, other_line)-> bool:
        """判断当前直线是否与另一条直线平行"""
        return self.slope() == other_line.slope()

    def perpendicular_to(self, other_line) -> bool:
        """判断当前直线是否与另一条直线垂直（正交）"""
        return self.slope() * other_line.slope() == -1

    def is_perpendicular(self, other_line) -> bool:
        """判断当前直线是否与另一条直线近似垂直（正交）"""
        return abs(self.slope() * other_line.slope() + 1) < 1e-9

    def is_parallel(self, other_line) -> bool:
        """判断当前直线是否与另一条直线近似平行"""
        return abs(self.slope() - other_line.slope()) < 1e-9

    def angle_with_x_axis(self) -> float:
        """计算当前直线与 x 轴的夹角"""
        return math.degrees(math.atan(self.slope()))

    def angle_with_another_line(self, other_line)->float:
        """计算当前直线与另一条直线的夹角"""
        if self.is_parallel(other_line):
            return 0.0
        angle = abs(math.atan((other_line.slope() - self.slope()) / (1 + self.slope() * other_line.slope())))
        return math.degrees(angle)

    def as_tuple(self):
        """将当前直线表示为一个元组，包含两个点的坐标元组"""
        return (self.point1.as_tuple(), self.point2.as_tuple())

    def as_numpy(self):
        """
        GraphLine类型转为numpy类型
        >>> point1 = GraphPoint(1, 2)
        >>> point2 = GraphPoint(3, 4)
        >>> line = GraphLine(point1, point2)
        >>> point_numpy = line.as_numpy()
        >>> # (array([1, 2]), array([3, 4]))
        """
        if self.point1.z is not None and self.point2.z is not None:
            point1 = np.array([self.point1.x, self.point1.y, self.point1.z])
            point2 = np.array([self.point2.x, self.point2.y, self.point2.z])
        else:
            point1 = np.array([self.point1.x, self.point1.y])
            point2 = np.array([self.point2.x, self.point2.y])
        return point1, point2

    @classmethod
    def from_numpy(cls, np_array):
        """
        numpy转为GraphLine类型
        >>> point1_numpy = np.array([1, 2])
        >>> point2_numpy = np.array([3, 4])
        >>> line = GraphLine.from_numpy((point1_numpy, point2_numpy))
        >>> ((1, 2), (3, 4))
        """
        point1_numpy = np.array(np_array[0])
        point2_numpy = np.array(np_array[1])

        if point1_numpy.shape == (2,):
            return cls(GraphPoint(point1_numpy[0], point1_numpy[1]), GraphPoint(point2_numpy[0], point2_numpy[1]))
        elif point1_numpy.shape == (3,):
            return cls(
                GraphPoint(point1_numpy[0], point1_numpy[1], point1_numpy[2]),
                GraphPoint(point2_numpy[0], point2_numpy[1], point2_numpy[2])
            )
        else:
            raise ValueError("Invalid array shape for conversion.")


if __name__ == "__main__":
    point1_numpy = np.array([1, 2])
    point2_numpy = np.array([3, 4])
    line = GraphLine.from_numpy((point1_numpy, point2_numpy))

    print(line)


"""
Original paper addresshttps: https://arxiv.org/pdf/1709.01507.pdf
Code originates from https://github.com/moskomule/senet.pytorch/blob/master/senet/se_module.py
Blog records: https://blog.csdn.net/m0_62919535/article/details/135761713
Time: 2024-01-23
"""
import torch
from torch import nn

class SELayer(nn.Module):
    def __init__(self, channel, reduction=16):
        super().__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(channel, channel // reduction, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(channel // reduction, channel, bias=False),
            nn.Sigmoid()
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y.expand_as(x)


if __name__ == '__main__':
    input = torch.randn(1, 64, 64, 64)
    se = SELayer(channel=64, reduction=8)
    output = se(input)
    print(output.shape)



from .savemodels import *
from .OneHot import *
from .learnrate import *
from .loss_function import *
from .loss_utils import *
from .avgweight import AveragingBaseModel, EMAModel, SWAModel, T_ADEMAModel



"""
论文原址：https://arxiv.org/pdf/2307.08388v2.pdf
According to https://github.com/YaoleiQi/DSCNet Newly submitted S3_DSConv_pro.py file was rewritten
because it used einops, so I changed everything involved to Torch handwriting implementation.
"""
import torch
import torch.nn as nn
import torch.nn.functional as F

class DSConv(nn.Module):
    def __init__(self, in_channels=1, out_channels=1, kernel_size=9, extend_scope=1.0,
                 morph=0, if_offset=True, device="cuda",):
        """
        :param in_channels: 输入通道的数量
        :param out_channels: 输出通道的数量
        :param kernel_size: 内核的大小
        :param extend_scope: 要扩大的范围。此方法的默认值为 1
        :param morph: 卷积核的形态主要沿 x轴（0）和 y轴（1）分为两类
        :param if_offset: 是否需要变形，如果为False，则为标准卷积核。
        :param device: 设备，'cpu' 或 'cuda'
        """
        super().__init__()
        self.kernel_size = kernel_size
        self.extend_scope = extend_scope
        if morph in (0, 1):
            self.morph = morph
        else:
            raise ValueError("[DSConv]:Only two types of morph: 0 (x-axis) or 1 (y-axis)")
        self.if_offset = if_offset
        self.device = device

        # self.bn = nn.BatchNorm2d(2 * kernel_size)
        self.gn_offset = nn.GroupNorm(kernel_size, 2 * kernel_size)
        self.gn = nn.GroupNorm(out_channels // 4, out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.tanh = nn.Tanh()

        self.offset_conv = nn.Conv2d(in_channels, 2 * kernel_size, 3, padding=1)

        self.dsc_conv_x = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=(kernel_size, 1),
            stride=(kernel_size, 1),
            padding=0,
        )
        self.dsc_conv_y = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=(1, kernel_size),
            stride=(1, kernel_size),
            padding=0,
        )

    def forward(self, input):
        offset = self.offset_conv(input)
        offset = self.gn_offset(offset)
        offset = self.tanh(offset)

        y_coordinate_map, x_coordinate_map = get_coordinate_map_2D(
            offset=offset,
            morph=self.morph,
            extend_scope=self.extend_scope,
            device=self.device,
        )
        deformed_feature = get_interpolated_feature(
            input,
            y_coordinate_map,
            x_coordinate_map,
        )
        output = None
        if self.morph == 0:
            output = self.dsc_conv_x(deformed_feature)
        elif self.morph == 1:
            output = self.dsc_conv_y(deformed_feature)

        # Groupnorm & ReLU
        output = self.gn(output)
        output = self.relu(output)

        return output

def _offset_to_center(kernel_size, width, height, device):
    """
    将偏移量转换为中心坐标
    :param kernel_size: 卷积核尺寸
    :param width: 输入图像的宽度
    :param height: 输入图像的高度
    :param device: 设备类型
    :return: _x_center: 转换后的 x 中心坐标，形状为 [kernel_size, width, height]
             _y_center: 转换后的 y 中心坐标，形状为 [kernel_size, width, height]
    """
    y_center = torch.arange(0, width, dtype=torch.float32, device=device)
    y_center = y_center.unsqueeze(0).unsqueeze(2)
    _y_center = y_center.repeat(kernel_size, 1, height)
    x_center = torch.arange(0, height, dtype=torch.float32, device=device)
    x_center_expanded = x_center.unsqueeze(0)
    _x_center = x_center_expanded.repeat(kernel_size, width, 1)
    return _x_center, _y_center

def get_coordinate_map_2D(offset, morph, extend_scope=1.0, device="cuda"):
    """
    :param offset: 通过形状为[B，2*K，W，H]的网络预测偏移量。这里的K是指内核大小。
    :param morph: 卷积核的形态主要沿 x轴（0）和 y轴（1）分为两类
    :param extend_scope: 要扩大的范围。此方法的默认值为1
    :param device: 设备，'cpu' 或 'cuda'
    :return: y_coordinate_map: coordinate map along y-axis with shape [B, K_H * H, K_W * W]
             x_coordinate_map: coordinate map along x-axis with shape [B, K_H * H, K_W * W]
    """
    batch_size, _, width, height = offset.shape
    kernel_size = offset.shape[1] // 2
    center = kernel_size // 2
    device = torch.device(device)
    y_offset_, x_offset_ = torch.split(offset, kernel_size, dim=1)
    x_center, y_center = _offset_to_center(kernel_size,width,height,device)

    if morph == 0:
        """
        Initialize the kernel and flatten the kernel
            y: only need 0
            x: -num_points//2 ~ num_points//2 (Determined by the kernel size)
        """
        y_spread_ = torch.zeros([kernel_size], device=device)
        x_spread_ = torch.linspace(-center, center, kernel_size, device=device)

        y_spread_expanded = y_spread_.unsqueeze(-1).unsqueeze(-1)
        y_grid_ = y_spread_expanded.repeat(1, width, height)
        x_spread_expanded = x_spread_.unsqueeze(-1).unsqueeze(-1)
        x_grid_ = x_spread_expanded.repeat(1, width, height)

        y_new_ = y_center + y_grid_
        x_new_ = x_center + x_grid_

        y_new_ = y_new_.repeat(batch_size, 1, 1, 1)
        x_new_ = x_new_.repeat(batch_size, 1, 1, 1)

        y_offset_ = y_offset_.permute(1, 0, 2, 3)
        y_offset_new_ = y_offset_.detach().clone()

        # 中心位置保持不变，其余位置开始摆动
        # 这部分很简单。其主要思想是“偏移是一个迭代过程”

        y_offset_new_[center] = 0
        for index in range(1, center + 1):
            y_offset_new_[center + index] = (
                    y_offset_new_[center + index - 1] + y_offset_[center + index]
            )
            y_offset_new_[center - index] = (
                    y_offset_new_[center - index + 1] + y_offset_[center - index]
            )

        y_offset_new_ = y_offset_new_.permute(1, 0, 2, 3)
        y_new_ = y_new_.add(y_offset_new_.mul(extend_scope))

        y_coordinate_map = y_new_.permute(0, 2, 1, 3).reshape(batch_size, -1, height)
        x_coordinate_map = x_new_.permute(0, 2, 1, 3).reshape(batch_size, -1, height)
        return y_coordinate_map, x_coordinate_map
    elif morph == 1:
        """
        Initialize the kernel and flatten the kernel
            y: -num_points//2 ~ num_points//2 (Determined by the kernel size)
            x: only need 0
        """
        y_spread_ = torch.linspace(-center, center, kernel_size, device=device)
        x_spread_ = torch.zeros([kernel_size], device=device)

        y_spread_expanded = y_spread_.unsqueeze(-1).unsqueeze(-1)
        y_grid_ = y_spread_expanded.repeat(1, width, height)
        x_spread_expanded = x_spread_.unsqueeze(-1).unsqueeze(-1)
        x_grid_ = x_spread_expanded.repeat(1, width, height)

        y_new_ = x_center + y_grid_
        x_new_ = x_center + x_grid_

        y_new_ = y_new_.repeat(batch_size, 1, 1, 1)
        x_new_ = x_new_.repeat(batch_size, 1, 1, 1)

        x_offset_ = x_offset_.permute(1, 0, 2, 3)
        x_offset_new_ = x_offset_.detach().clone()

        x_offset_new_[center] = 0
        for index in range(1, center + 1):
            x_offset_new_[center + index] = (
                    x_offset_new_[center + index - 1] + x_offset_[center + index]
            )
            x_offset_new_[center - index] = (
                    x_offset_new_[center - index + 1] + x_offset_[center - index]
            )

        x_offset_new_ = x_offset_new_.permute(1, 0, 2, 3)
        x_new_ = x_new_.add(x_offset_new_.mul(extend_scope))
        y_coordinate_map = y_new_.permute(0, 2, 3, 1).reshape(batch_size, width, -1)
        x_coordinate_map = x_new_.permute(0, 2, 3, 1).reshape(batch_size, width, -1)

        return y_coordinate_map, x_coordinate_map


def get_interpolated_feature(input_feature, y_coordinate_map, x_coordinate_map, interpolate_mode="bilinear",):
    """
    根据坐标映射，对 DSCNet 的特征进行插值，插值后的特征，形状为 [B, C, K_H * H, K_W * W]
    :param input_feature: 需要插值的特征，形状为 [B, C, H, W]
    :param y_coordinate_map: 沿 y 轴的坐标映射，形状为 [B, K_H * H, K_W * W]
    :param x_coordinate_map: 沿 x 轴的坐标映射，形状为 [B, K_H * H, K_W * W]
    :param interpolate_mode: F.grid_sample 的 'mode' 参数，可以是 'bilinear' 或 'bicubic'。默认为 'bilinear'。
    """
    if interpolate_mode not in ("bilinear", "bicubic"):
        raise ValueError("interpolate_mode should be 'bilinear' or 'bicubic'.")

    y_max = input_feature.shape[-2] - 1
    x_max = input_feature.shape[-1] - 1

    y_coordinate_map_ = _coordinate_map_scaling(y_coordinate_map, origin=[0, y_max])
    x_coordinate_map_ = _coordinate_map_scaling(x_coordinate_map, origin=[0, x_max])

    y_coordinate_map_ = torch.unsqueeze(y_coordinate_map_, dim=-1)
    x_coordinate_map_ = torch.unsqueeze(x_coordinate_map_, dim=-1)

    # Note here grid with shape [B, H, W, 2]
    # Where [:, :, :, 2] refers to [x ,y]
    grid = torch.cat([x_coordinate_map_, y_coordinate_map_], dim=-1)

    interpolated_feature = F.grid_sample(
        input=input_feature,
        grid=grid,
        mode=interpolate_mode,
        padding_mode="zeros",
        align_corners=True,
    )

    return interpolated_feature


def _coordinate_map_scaling(coordinate_map, origin, target=(-1, 1)):
    """
    将坐标映射尺度进行缩放
    :param coordinate_map: 坐标映射张量
    :param origin: 原始尺度范围，e.g.[coordinate_map.min(), coordinate_map.max()]
    :param target: 目标尺度范围，默认为 (-1, 1)
    """
    min, max = origin
    a, b = target
    coordinate_map_scaled = torch.clamp(coordinate_map, min, max)
    scale_factor = (b - a) / (max - min)
    coordinate_map_scaled = scale_factor * (coordinate_map_scaled - min)  # 转成[0, max-min]的范围，缩放到 [0, (b-a)]
    coordinate_map_scaled = coordinate_map_scaled + a  # 平移到目标尺度范围 [a, b]

    return coordinate_map_scaled


if __name__=="__main__":
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    A = torch.tensor([[[[0.1, 0.2, 0.3],
                        [0.4, 0.5, 0.6],
                        [0.7, 0.8, 0.9]]],

                      [[[1.0, 1.1, 1.2],
                        [1.3, 1.4, 1.5],
                        [1.6, 1.7, 1.8]]],

                      [[[2.0, 2.1, 2.2],
                        [2.3, 2.4, 2.5],
                        [2.6, 2.7, 2.8]]],

                      [[[3.0, 3.1, 3.2],
                        [3.3, 3.4, 3.5],
                        [3.6, 3.7, 3.8]]]])

    print("输入形状", A.shape)  # torch.Size([4, 1, 3, 3])
    conv0 = DSConv(
        in_channels=1,
        out_channels=10,
        kernel_size=15,
        extend_scope=1,
        morph=0,
        if_offset=True,
        device="cpu")
    if torch.cuda.is_available():
        A = A.to(device)
        conv0 = conv0.to(device)
    out = conv0(A)
    print("输出形状：",out.shape)   # torch.Size([4, 10, 3, 3])
    # print(out)


import math

def autopad(kernel, padding=None, dilation=1):
    """自动计算填充大小，以使输出具有与输入相同的形状
    :param k: kernel
    :param p: padding
    :param d: dilation
    :return: 自动计算得到的填充大小
    """
    k, p, d = kernel, padding, dilation
    if d > 1:
        k = d * (k - 1) + 1 if isinstance(k, int) else [d * (x - 1) + 1 for x in k]  # actual kernel-size
    if p is None:
        p = k // 2 if isinstance(k, int) else [x // 2 for x in k]  # auto-pad
    return p

def _make_divisible(v, divisor, min_value=None):
    """
    This function is taken from the original tf repo.It can be seen here:
        https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    Args:
        v: The number of input channels.
        divisor: The number of channels should be a multiple of this value.
        min_value: The minimum value of the number of channels, which defaults to the advisor.

    Returns: It ensures that all layers have a channel number that is divisible by 8
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v

def convpool_outsize(input_size, kernel_size, stride, padding, is_conv=True, ceil_mode=False):
    """
    计算卷积或池化的输出尺寸。
    Parameters:
        input_size (tuple): 输入尺寸，格式为 (height, width, channels)
        kernel_size (int or tuple): 卷积核或池化核的尺寸，如果是 int，表示高宽相同的正方形核
        stride (int): 步幅
        padding (int): 填充大小
        is_conv (bool): 是否是卷积操作，如果是 True，使用卷积输出尺寸的计算公式；否则使用池化输出尺寸的计算公式

    Returns:
        output_size (tuple): 输出尺寸，格式为 (height, width, channels)。
    """
    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)

    h_in, w_in, c = input_size
    h_k, w_k = kernel_size

    if is_conv:
        h_out = ((h_in + 2 * padding - h_k) // stride) + 1
        w_out = ((w_in + 2 * padding - w_k) // stride) + 1
    else:
        h_out = ((h_in - h_k) // stride) + 1
        w_out = ((w_in - w_k) // stride) + 1
    if ceil_mode:
        h_out, w_out = math.ceil(h_out), math.ceil(w_out)
    return h_out, w_out, c

if __name__=="__main__":
    input_size = (112, 112, 64)
    kernel_size = 3
    stride = 2
    padding = 3
    conv_output_size = convpool_outsize(input_size, kernel_size, stride, padding, is_conv=True)
    print("卷积输出尺寸:", conv_output_size)
    pool_output_size = convpool_outsize(input_size, kernel_size, stride, padding, is_conv=False)
    print("池化输出尺寸:", pool_output_size)



"""

Original paper addresshttps: https://arxiv.org/pdf/1807.06521.pdf
"""

import torch
from torch import nn

class ChannelAttention(nn.Module):
    def __init__(self, in_planes, ratio=16):
        super(ChannelAttention, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.max_pool = nn.AdaptiveMaxPool2d(1)

        self.fc1 = nn.Conv2d(in_planes, in_planes // ratio, 1, bias=False)
        self.relu1 = nn.ReLU()
        self.fc2 = nn.Conv2d(in_planes // ratio, in_planes, 1, bias=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = self.fc2(self.relu1(self.fc1(self.avg_pool(x))))
        max_out = self.fc2(self.relu1(self.fc1(self.max_pool(x))))
        out = avg_out + max_out
        return self.sigmoid(out)


class SpatialAttention(nn.Module):
    def __init__(self, kernel_size=7):
        super(SpatialAttention, self).__init__()

        assert kernel_size in (3, 7), 'kernel size must be 3 or 7'
        padding = 3 if kernel_size == 7 else 1

        self.conv1 = nn.Conv2d(2, 1, kernel_size, padding=padding, bias=False)  # 7,3     3,1
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = torch.mean(x, dim=1, keepdim=True)
        max_out, _ = torch.max(x, dim=1, keepdim=True)
        x = torch.cat([avg_out, max_out], dim=1)
        x = self.conv1(x)
        return self.sigmoid(x)


class CBAM(nn.Module):
    def __init__(self, in_planes, ratio=16, kernel_size=7):
        super(CBAM, self).__init__()
        self.ca = ChannelAttention(in_planes, ratio)
        self.sa = SpatialAttention(kernel_size)

    def forward(self, x):
        out = x * self.ca(x)
        result = out * self.sa(out)
        return result



if __name__ == '__main__':
    block = CBAM(64)
    input = torch.rand(1, 64, 64, 64)
    output = block(input)
    print(input.size(), output.size())